from typing import Any, Dict, Type, TypeVar, TYPE_CHECKING

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

from typing import Literal
from typing import cast, Union
from typing import cast, List
from ..types import UNSET, Unset
from typing import cast
from typing import Dict
from typing import Union

if TYPE_CHECKING:
    from ..models.validation_error_response_part import ValidationErrorResponsePart


T = TypeVar("T", bound="ValidationErrorResponse")


@_attrs_define
class ValidationErrorResponse:
    """
    Attributes:
        status (int):
        type (Literal['urn:invariant:errors:validation']):
        title (Literal['There was a problem with your request.']):
        detail (str):
        validations (List['ValidationErrorResponsePart']):
        instance (Union[None, Unset, str]):
    """

    status: int
    type: Literal["urn:invariant:errors:validation"]
    title: Literal["There was a problem with your request."]
    detail: str
    validations: List["ValidationErrorResponsePart"]
    instance: Union[None, Unset, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        status = self.status
        type = self.type
        title = self.title
        detail = self.detail
        validations = []
        for validations_item_data in self.validations:
            validations_item = validations_item_data.to_dict()

            validations.append(validations_item)

        instance: Union[None, Unset, str]
        if isinstance(self.instance, Unset):
            instance = UNSET

        else:
            instance = self.instance

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "status": status,
                "type": type,
                "title": title,
                "detail": detail,
                "validations": validations,
            }
        )
        if instance is not UNSET:
            field_dict["instance"] = instance

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.validation_error_response_part import ValidationErrorResponsePart

        d = src_dict.copy()
        status = d.pop("status")

        type = d.pop("type")

        title = d.pop("title")

        detail = d.pop("detail")

        validations = []
        _validations = d.pop("validations")
        for validations_item_data in _validations:
            validations_item = ValidationErrorResponsePart.from_dict(
                validations_item_data
            )

            validations.append(validations_item)

        def _parse_instance(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        instance = _parse_instance(d.pop("instance", UNSET))

        validation_error_response = cls(
            status=status,
            type=type,
            title=title,
            detail=detail,
            validations=validations,
            instance=instance,
        )

        validation_error_response.additional_properties = d
        return validation_error_response

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, Type, TypeVar, TYPE_CHECKING

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field


from typing import cast, Union
from typing import cast
from typing import Dict

if TYPE_CHECKING:
    from ..models.user import User
    from ..models.organization import Organization
    from ..models.user_tabs_config import UserTabsConfig


T = TypeVar("T", bound="UIStatusResponse")


@_attrs_define
class UIStatusResponse:
    """
    Attributes:
        user (User):
        organization (Organization): The internal model inside the database.
        tabs (Union['UserTabsConfig', None]):
    """

    user: "User"
    organization: "Organization"
    tabs: Union["UserTabsConfig", None]
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        from ..models.user_tabs_config import UserTabsConfig

        user = self.user.to_dict()

        organization = self.organization.to_dict()

        tabs: Union[Dict[str, Any], None]

        if isinstance(self.tabs, UserTabsConfig):
            tabs = self.tabs.to_dict()

        else:
            tabs = self.tabs

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "user": user,
                "organization": organization,
                "tabs": tabs,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.user import User
        from ..models.organization import Organization
        from ..models.user_tabs_config import UserTabsConfig

        d = src_dict.copy()
        user = User.from_dict(d.pop("user"))

        organization = Organization.from_dict(d.pop("organization"))

        def _parse_tabs(data: object) -> Union["UserTabsConfig", None]:
            if data is None:
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                tabs_type_0 = UserTabsConfig.from_dict(data)

                return tabs_type_0
            except:  # noqa: E722
                pass
            return cast(Union["UserTabsConfig", None], data)

        tabs = _parse_tabs(d.pop("tabs"))

        ui_status_response = cls(
            user=user,
            organization=organization,
            tabs=tabs,
        )

        ui_status_response.additional_properties = d
        return ui_status_response

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, Type, TypeVar, TYPE_CHECKING

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field


import datetime
from dateutil.parser import isoparse
from typing import Dict

if TYPE_CHECKING:
    from ..models.external_status_data_integration import ExternalStatusDataIntegration


T = TypeVar("T", bound="ExternalStatusIntegration")


@_attrs_define
class ExternalStatusIntegration:
    """
    Attributes:
        organization_uuid (str):
        subject_uuid (str):
        data (ExternalStatusDataIntegration):
        created_at (datetime.datetime):
    """

    organization_uuid: str
    subject_uuid: str
    data: "ExternalStatusDataIntegration"
    created_at: datetime.datetime
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        organization_uuid = self.organization_uuid
        subject_uuid = self.subject_uuid
        data = self.data.to_dict()

        created_at = self.created_at.isoformat()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "organization_uuid": organization_uuid,
                "subject_uuid": subject_uuid,
                "data": data,
                "created_at": created_at,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.external_status_data_integration import (
            ExternalStatusDataIntegration,
        )

        d = src_dict.copy()
        organization_uuid = d.pop("organization_uuid")

        subject_uuid = d.pop("subject_uuid")

        data = ExternalStatusDataIntegration.from_dict(d.pop("data"))

        created_at = isoparse(d.pop("created_at"))

        external_status_integration = cls(
            organization_uuid=organization_uuid,
            subject_uuid=subject_uuid,
            data=data,
            created_at=created_at,
        )

        external_status_integration.additional_properties = d
        return external_status_integration

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, Type, TypeVar

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field


from typing import cast, Union
from ..models.new_password_request_authn_type import NewPasswordRequestAuthnType


T = TypeVar("T", bound="NewPasswordRequest")


@_attrs_define
class NewPasswordRequest:
    """
    Attributes:
        email (Union[None, str]):
        password (str):
        authn_type (NewPasswordRequestAuthnType):
        authn (str):
    """

    email: Union[None, str]
    password: str
    authn_type: NewPasswordRequestAuthnType
    authn: str
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        email: Union[None, str]

        email = self.email

        password = self.password
        authn_type = self.authn_type.value

        authn = self.authn

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "email": email,
                "password": password,
                "authn_type": authn_type,
                "authn": authn,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()

        def _parse_email(data: object) -> Union[None, str]:
            if data is None:
                return data
            return cast(Union[None, str], data)

        email = _parse_email(d.pop("email"))

        password = d.pop("password")

        authn_type = NewPasswordRequestAuthnType(d.pop("authn_type"))

        authn = d.pop("authn")

        new_password_request = cls(
            email=email,
            password=password,
            authn_type=authn_type,
            authn=authn,
        )

        new_password_request.additional_properties = d
        return new_password_request

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class ChallengeResponseChallenge(str, Enum):
    BASIC_AUTH = "basic_auth"
    NEW_PASSWORD = "new_password"
    RESET_PIN = "reset_pin"
    VALIDATE_EMAIL = "validate_email"

    def __str__(self) -> str:
        return str(self.value)



from typing import Any, Dict, Type, TypeVar

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

from ..types import UNSET, Unset
from typing import Union


T = TypeVar("T", bound="ReportTextSummaryRequest")


@_attrs_define
class ReportTextSummaryRequest:
    """
    Attributes:
        mode (Union[Unset, str]):  Default: 'text'.
        traces (Union[Unset, bool]):
    """

    mode: Union[Unset, str] = "text"
    traces: Union[Unset, bool] = False
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        mode = self.mode
        traces = self.traces

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if mode is not UNSET:
            field_dict["mode"] = mode
        if traces is not UNSET:
            field_dict["traces"] = traces

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        mode = d.pop("mode", UNSET)

        traces = d.pop("traces", UNSET)

        report_text_summary_request = cls(
            mode=mode,
            traces=traces,
        )

        report_text_summary_request.additional_properties = d
        return report_text_summary_request

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, Type, TypeVar, TYPE_CHECKING

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

from typing import cast, Union
from ..types import UNSET, Unset
from typing import cast
from typing import Dict
from typing import Union

if TYPE_CHECKING:
    from ..models.email_password_login import EmailPasswordLogin
    from ..models.validation_request import ValidationRequest
    from ..models.reset_pin_request import ResetPINRequest
    from ..models.new_password_request import NewPasswordRequest
    from ..models.reset_request import ResetRequest


T = TypeVar("T", bound="LoginRequest")


@_attrs_define
class LoginRequest:
    """Request to determine login method (Basic, SAML, etc).

    Attributes:
        basic_auth (Union['EmailPasswordLogin', None, Unset]):
        email_valid (Union['ValidationRequest', None, Unset]):
        reset_request (Union['ResetRequest', None, Unset]):
        reset_pin_request (Union['ResetPINRequest', None, Unset]):
        new_password (Union['NewPasswordRequest', None, Unset]):
    """

    basic_auth: Union["EmailPasswordLogin", None, Unset] = UNSET
    email_valid: Union["ValidationRequest", None, Unset] = UNSET
    reset_request: Union["ResetRequest", None, Unset] = UNSET
    reset_pin_request: Union["ResetPINRequest", None, Unset] = UNSET
    new_password: Union["NewPasswordRequest", None, Unset] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        from ..models.email_password_login import EmailPasswordLogin
        from ..models.validation_request import ValidationRequest
        from ..models.reset_pin_request import ResetPINRequest
        from ..models.new_password_request import NewPasswordRequest
        from ..models.reset_request import ResetRequest

        basic_auth: Union[Dict[str, Any], None, Unset]
        if isinstance(self.basic_auth, Unset):
            basic_auth = UNSET

        elif isinstance(self.basic_auth, EmailPasswordLogin):
            basic_auth = UNSET
            if not isinstance(self.basic_auth, Unset):
                basic_auth = self.basic_auth.to_dict()

        else:
            basic_auth = self.basic_auth

        email_valid: Union[Dict[str, Any], None, Unset]
        if isinstance(self.email_valid, Unset):
            email_valid = UNSET

        elif isinstance(self.email_valid, ValidationRequest):
            email_valid = UNSET
            if not isinstance(self.email_valid, Unset):
                email_valid = self.email_valid.to_dict()

        else:
            email_valid = self.email_valid

        reset_request: Union[Dict[str, Any], None, Unset]
        if isinstance(self.reset_request, Unset):
            reset_request = UNSET

        elif isinstance(self.reset_request, ResetRequest):
            reset_request = UNSET
            if not isinstance(self.reset_request, Unset):
                reset_request = self.reset_request.to_dict()

        else:
            reset_request = self.reset_request

        reset_pin_request: Union[Dict[str, Any], None, Unset]
        if isinstance(self.reset_pin_request, Unset):
            reset_pin_request = UNSET

        elif isinstance(self.reset_pin_request, ResetPINRequest):
            reset_pin_request = UNSET
            if not isinstance(self.reset_pin_request, Unset):
                reset_pin_request = self.reset_pin_request.to_dict()

        else:
            reset_pin_request = self.reset_pin_request

        new_password: Union[Dict[str, Any], None, Unset]
        if isinstance(self.new_password, Unset):
            new_password = UNSET

        elif isinstance(self.new_password, NewPasswordRequest):
            new_password = UNSET
            if not isinstance(self.new_password, Unset):
                new_password = self.new_password.to_dict()

        else:
            new_password = self.new_password

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if basic_auth is not UNSET:
            field_dict["basic_auth"] = basic_auth
        if email_valid is not UNSET:
            field_dict["email_valid"] = email_valid
        if reset_request is not UNSET:
            field_dict["reset_request"] = reset_request
        if reset_pin_request is not UNSET:
            field_dict["reset_pin_request"] = reset_pin_request
        if new_password is not UNSET:
            field_dict["new_password"] = new_password

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.email_password_login import EmailPasswordLogin
        from ..models.validation_request import ValidationRequest
        from ..models.reset_pin_request import ResetPINRequest
        from ..models.new_password_request import NewPasswordRequest
        from ..models.reset_request import ResetRequest

        d = src_dict.copy()

        def _parse_basic_auth(data: object) -> Union["EmailPasswordLogin", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                _basic_auth_type_0 = data
                basic_auth_type_0: Union[Unset, EmailPasswordLogin]
                if isinstance(_basic_auth_type_0, Unset):
                    basic_auth_type_0 = UNSET
                else:
                    basic_auth_type_0 = EmailPasswordLogin.from_dict(_basic_auth_type_0)

                return basic_auth_type_0
            except:  # noqa: E722
                pass
            return cast(Union["EmailPasswordLogin", None, Unset], data)

        basic_auth = _parse_basic_auth(d.pop("basic_auth", UNSET))

        def _parse_email_valid(data: object) -> Union["ValidationRequest", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                _email_valid_type_0 = data
                email_valid_type_0: Union[Unset, ValidationRequest]
                if isinstance(_email_valid_type_0, Unset):
                    email_valid_type_0 = UNSET
                else:
                    email_valid_type_0 = ValidationRequest.from_dict(
                        _email_valid_type_0
                    )

                return email_valid_type_0
            except:  # noqa: E722
                pass
            return cast(Union["ValidationRequest", None, Unset], data)

        email_valid = _parse_email_valid(d.pop("email_valid", UNSET))

        def _parse_reset_request(data: object) -> Union["ResetRequest", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                _reset_request_type_0 = data
                reset_request_type_0: Union[Unset, ResetRequest]
                if isinstance(_reset_request_type_0, Unset):
                    reset_request_type_0 = UNSET
                else:
                    reset_request_type_0 = ResetRequest.from_dict(_reset_request_type_0)

                return reset_request_type_0
            except:  # noqa: E722
                pass
            return cast(Union["ResetRequest", None, Unset], data)

        reset_request = _parse_reset_request(d.pop("reset_request", UNSET))

        def _parse_reset_pin_request(
            data: object,
        ) -> Union["ResetPINRequest", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                _reset_pin_request_type_0 = data
                reset_pin_request_type_0: Union[Unset, ResetPINRequest]
                if isinstance(_reset_pin_request_type_0, Unset):
                    reset_pin_request_type_0 = UNSET
                else:
                    reset_pin_request_type_0 = ResetPINRequest.from_dict(
                        _reset_pin_request_type_0
                    )

                return reset_pin_request_type_0
            except:  # noqa: E722
                pass
            return cast(Union["ResetPINRequest", None, Unset], data)

        reset_pin_request = _parse_reset_pin_request(d.pop("reset_pin_request", UNSET))

        def _parse_new_password(
            data: object,
        ) -> Union["NewPasswordRequest", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                _new_password_type_0 = data
                new_password_type_0: Union[Unset, NewPasswordRequest]
                if isinstance(_new_password_type_0, Unset):
                    new_password_type_0 = UNSET
                else:
                    new_password_type_0 = NewPasswordRequest.from_dict(
                        _new_password_type_0
                    )

                return new_password_type_0
            except:  # noqa: E722
                pass
            return cast(Union["NewPasswordRequest", None, Unset], data)

        new_password = _parse_new_password(d.pop("new_password", UNSET))

        login_request = cls(
            basic_auth=basic_auth,
            email_valid=email_valid,
            reset_request=reset_request,
            reset_pin_request=reset_pin_request,
            new_password=new_password,
        )

        login_request.additional_properties = d
        return login_request

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, Type, TypeVar, TYPE_CHECKING

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field


from typing import cast, Union
from typing import cast
from typing import Literal
from ..models.generic_state import GenericState
from dateutil.parser import isoparse
from typing import Dict
import datetime

if TYPE_CHECKING:
    from ..models.base_error_response import BaseErrorResponse


T = TypeVar("T", bound="ExternalStatusDataIntegration")


@_attrs_define
class ExternalStatusDataIntegration:
    """
    Attributes:
        type (Literal['integration']):
        state (GenericState):
        error (Union['BaseErrorResponse', None]):
        last_used_at (datetime.datetime):
        modified_at (datetime.datetime):
    """

    type: Literal["integration"]
    state: GenericState
    error: Union["BaseErrorResponse", None]
    last_used_at: datetime.datetime
    modified_at: datetime.datetime
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        from ..models.base_error_response import BaseErrorResponse

        type = self.type
        state = self.state.value

        error: Union[Dict[str, Any], None]

        if isinstance(self.error, BaseErrorResponse):
            error = self.error.to_dict()

        else:
            error = self.error

        last_used_at = self.last_used_at.isoformat()

        modified_at = self.modified_at.isoformat()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "type": type,
                "state": state,
                "error": error,
                "last_used_at": last_used_at,
                "modified_at": modified_at,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.base_error_response import BaseErrorResponse

        d = src_dict.copy()
        type = d.pop("type")

        state = GenericState(d.pop("state"))

        def _parse_error(data: object) -> Union["BaseErrorResponse", None]:
            if data is None:
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                error_type_0 = BaseErrorResponse.from_dict(data)

                return error_type_0
            except:  # noqa: E722
                pass
            return cast(Union["BaseErrorResponse", None], data)

        error = _parse_error(d.pop("error"))

        last_used_at = isoparse(d.pop("last_used_at"))

        modified_at = isoparse(d.pop("modified_at"))

        external_status_data_integration = cls(
            type=type,
            state=state,
            error=error,
            last_used_at=last_used_at,
            modified_at=modified_at,
        )

        external_status_data_integration.additional_properties = d
        return external_status_data_integration

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, Type, TypeVar, TYPE_CHECKING

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field


import datetime
from dateutil.parser import isoparse
from typing import Dict

if TYPE_CHECKING:
    from ..models.metadata import Metadata


T = TypeVar("T", bound="Organization")


@_attrs_define
class Organization:
    """The internal model inside the database.

    Attributes:
        uuid (str):
        name (str):
        description (str):
        metadata (Metadata):
        is_active (bool):
        created_at (datetime.datetime):
    """

    uuid: str
    name: str
    description: str
    metadata: "Metadata"
    is_active: bool
    created_at: datetime.datetime
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        uuid = self.uuid
        name = self.name
        description = self.description
        metadata = self.metadata.to_dict()

        is_active = self.is_active
        created_at = self.created_at.isoformat()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "uuid": uuid,
                "name": name,
                "description": description,
                "metadata": metadata,
                "is_active": is_active,
                "created_at": created_at,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.metadata import Metadata

        d = src_dict.copy()
        uuid = d.pop("uuid")

        name = d.pop("name")

        description = d.pop("description")

        metadata = Metadata.from_dict(d.pop("metadata"))

        is_active = d.pop("is_active")

        created_at = isoparse(d.pop("created_at"))

        organization = cls(
            uuid=uuid,
            name=name,
            description=description,
            metadata=metadata,
            is_active=is_active,
            created_at=created_at,
        )

        organization.additional_properties = d
        return organization

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, Type, TypeVar, TYPE_CHECKING

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field


from typing import Dict

if TYPE_CHECKING:
    from ..models.integration import Integration
    from ..models.external_status_integration import ExternalStatusIntegration


T = TypeVar("T", bound="IntegrationWithStatus")


@_attrs_define
class IntegrationWithStatus:
    """
    Attributes:
        integration (Integration):
        status (ExternalStatusIntegration):
    """

    integration: "Integration"
    status: "ExternalStatusIntegration"
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        integration = self.integration.to_dict()

        status = self.status.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "integration": integration,
                "status": status,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.integration import Integration
        from ..models.external_status_integration import ExternalStatusIntegration

        d = src_dict.copy()
        integration = Integration.from_dict(d.pop("integration"))

        status = ExternalStatusIntegration.from_dict(d.pop("status"))

        integration_with_status = cls(
            integration=integration,
            status=status,
        )

        integration_with_status.additional_properties = d
        return integration_with_status

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



""" Contains all the data models used in inputs/outputs """

from .base_error_response import BaseErrorResponse
from .body_upload_snapshot_organization_name_api_v1_uploadsnapshot_post import (
    BodyUploadSnapshotOrganizationNameApiV1UploadsnapshotPost,
)
from .challenge_response import ChallengeResponse
from .challenge_response_challenge import ChallengeResponseChallenge
from .comparison_reportdata import ComparisonReportdata
from .comparison_reportdata_files import ComparisonReportdataFiles
from .create_integration_request_github_app_installation import (
    CreateIntegrationRequestGithubAppInstallation,
)
from .create_integration_request_github_app_installation_data import (
    CreateIntegrationRequestGithubAppInstallationData,
)
from .create_token_request import CreateTokenRequest
from .external_status_data_integration import ExternalStatusDataIntegration
from .external_status_integration import ExternalStatusIntegration
from .generic_state import GenericState
from .get_report_summary_response import GetReportSummaryResponse
from .get_report_summary_response_status import GetReportSummaryResponseStatus
from .get_report_summary_response_summary import GetReportSummaryResponseSummary
from .integration import Integration
from .integration_data_github_app_installation import (
    IntegrationDataGithubAppInstallation,
)
from .integration_data_github_app_installation_data import (
    IntegrationDataGithubAppInstallationData,
)
from .integration_data_github_app_installation_data_extra import (
    IntegrationDataGithubAppInstallationDataExtra,
)
from .integration_with_status import IntegrationWithStatus
from .invite_user_request import InviteUserRequest
from .list_reports_response import ListReportsResponse
from .metadata import Metadata
from .organization import Organization
from .poc_report_data import POCReportData
from .refresh_response import RefreshResponse
from .report import Report
from .report_text_summary_request import ReportTextSummaryRequest
from .report_text_summary_response import ReportTextSummaryResponse
from .snapshot_report_data import SnapshotReportData
from .snapshot_report_data_files import SnapshotReportDataFiles
from .tab_info import TabInfo
from .tab_info_parameters_type_0 import TabInfoParametersType0
from .tab_info_state_type_0 import TabInfoStateType0
from .ui_status_response import UIStatusResponse
from .upload_snapshot_response import UploadSnapshotResponse
from .upload_snapshot_status_response import UploadSnapshotStatusResponse
from .user import User
from .user_metadata import UserMetadata
from .user_tabs_config import UserTabsConfig
from .validation_error_response import ValidationErrorResponse
from .validation_error_response_part import ValidationErrorResponsePart

__all__ = (
    "BaseErrorResponse",
    "BodyUploadSnapshotOrganizationNameApiV1UploadsnapshotPost",
    "ChallengeResponse",
    "ChallengeResponseChallenge",
    "ComparisonReportdata",
    "ComparisonReportdataFiles",
    "CreateIntegrationRequestGithubAppInstallation",
    "CreateIntegrationRequestGithubAppInstallationData",
    "CreateTokenRequest",
    "ExternalStatusDataIntegration",
    "ExternalStatusIntegration",
    "GenericState",
    "GetReportSummaryResponse",
    "GetReportSummaryResponseStatus",
    "GetReportSummaryResponseSummary",
    "Integration",
    "IntegrationDataGithubAppInstallation",
    "IntegrationDataGithubAppInstallationData",
    "IntegrationDataGithubAppInstallationDataExtra",
    "IntegrationWithStatus",
    "InviteUserRequest",
    "ListReportsResponse",
    "Metadata",
    "Organization",
    "POCReportData",
    "RefreshResponse",
    "Report",
    "ReportTextSummaryRequest",
    "ReportTextSummaryResponse",
    "SnapshotReportData",
    "SnapshotReportDataFiles",
    "TabInfo",
    "TabInfoParametersType0",
    "TabInfoStateType0",
    "UIStatusResponse",
    "UploadSnapshotResponse",
    "UploadSnapshotStatusResponse",
    "User",
    "UserMetadata",
    "UserTabsConfig",
    "ValidationErrorResponse",
    "ValidationErrorResponsePart",
)



from typing import Any, Dict, Type, TypeVar

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field


T = TypeVar("T", bound="POCReportData")


@_attrs_define
class POCReportData:
    """
    Attributes:
        issues (str):
        edges (str):
        routers (str):
        nodes (str):
        external_ports (str):
        rule_findings (str):
        connect_to (str):
    """

    issues: str
    edges: str
    routers: str
    nodes: str
    external_ports: str
    rule_findings: str
    connect_to: str
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        issues = self.issues
        edges = self.edges
        routers = self.routers
        nodes = self.nodes
        external_ports = self.external_ports
        rule_findings = self.rule_findings
        connect_to = self.connect_to

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "issues": issues,
                "edges": edges,
                "routers": routers,
                "nodes": nodes,
                "external_ports": external_ports,
                "rule_findings": rule_findings,
                "connectTo": connect_to,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        issues = d.pop("issues")

        edges = d.pop("edges")

        routers = d.pop("routers")

        nodes = d.pop("nodes")

        external_ports = d.pop("external_ports")

        rule_findings = d.pop("rule_findings")

        connect_to = d.pop("connectTo")

        poc_report_data = cls(
            issues=issues,
            edges=edges,
            routers=routers,
            nodes=nodes,
            external_ports=external_ports,
            rule_findings=rule_findings,
            connect_to=connect_to,
        )

        poc_report_data.additional_properties = d
        return poc_report_data

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, Type, TypeVar, TYPE_CHECKING

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field


from typing import Union
from dateutil.parser import isoparse
from typing import Dict
import datetime

if TYPE_CHECKING:
    from ..models.snapshot_report_data import SnapshotReportData
    from ..models.poc_report_data import POCReportData


T = TypeVar("T", bound="Report")


@_attrs_define
class Report:
    """
    Attributes:
        uuid (str):
        organization_uuid (str):
        reports (Union['POCReportData', 'SnapshotReportData']):
        created_at (datetime.datetime):
    """

    uuid: str
    organization_uuid: str
    reports: Union["POCReportData", "SnapshotReportData"]
    created_at: datetime.datetime
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        from ..models.snapshot_report_data import SnapshotReportData

        uuid = self.uuid
        organization_uuid = self.organization_uuid
        reports: Dict[str, Any]

        if isinstance(self.reports, SnapshotReportData):
            reports = self.reports.to_dict()

        else:
            reports = self.reports.to_dict()

        created_at = self.created_at.isoformat()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "uuid": uuid,
                "organization_uuid": organization_uuid,
                "reports": reports,
                "created_at": created_at,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.snapshot_report_data import SnapshotReportData
        from ..models.poc_report_data import POCReportData

        d = src_dict.copy()
        uuid = d.pop("uuid")

        organization_uuid = d.pop("organization_uuid")

        def _parse_reports(
            data: object,
        ) -> Union["POCReportData", "SnapshotReportData"]:
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                reports_type_0 = SnapshotReportData.from_dict(data)

                return reports_type_0
            except:  # noqa: E722
                pass
            if not isinstance(data, dict):
                raise TypeError()
            reports_type_1 = POCReportData.from_dict(data)

            return reports_type_1

        reports = _parse_reports(d.pop("reports"))

        created_at = isoparse(d.pop("created_at"))

        report = cls(
            uuid=uuid,
            organization_uuid=organization_uuid,
            reports=reports,
            created_at=created_at,
        )

        report.additional_properties = d
        return report

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, Type, TypeVar, TYPE_CHECKING

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field


from typing import List
from typing import Dict

if TYPE_CHECKING:
    from ..models.report import Report


T = TypeVar("T", bound="ListReportsResponse")


@_attrs_define
class ListReportsResponse:
    """
    Attributes:
        reports (List['Report']):
    """

    reports: List["Report"]
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        reports = []
        for reports_item_data in self.reports:
            reports_item = reports_item_data.to_dict()

            reports.append(reports_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "reports": reports,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.report import Report

        d = src_dict.copy()
        reports = []
        _reports = d.pop("reports")
        for reports_item_data in _reports:
            reports_item = Report.from_dict(reports_item_data)

            reports.append(reports_item)

        list_reports_response = cls(
            reports=reports,
        )

        list_reports_response.additional_properties = d
        return list_reports_response

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, Type, TypeVar

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

from typing import Literal
from typing import cast, Union
from ..types import UNSET, Unset
from ..models.challenge_response_challenge import ChallengeResponseChallenge
from typing import Union


T = TypeVar("T", bound="ChallengeResponse")


@_attrs_define
class ChallengeResponse:
    """
    Attributes:
        status (int):
        type (Literal['urn:invariant:errors:auth_challenge']):
        title (str):
        detail (str):
        challenge (ChallengeResponseChallenge):
        instance (Union[None, Unset, str]):
        login_token (Union[None, Unset, str]):
    """

    status: int
    type: Literal["urn:invariant:errors:auth_challenge"]
    title: str
    detail: str
    challenge: ChallengeResponseChallenge
    instance: Union[None, Unset, str] = UNSET
    login_token: Union[None, Unset, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        status = self.status
        type = self.type
        title = self.title
        detail = self.detail
        challenge = self.challenge.value

        instance: Union[None, Unset, str]
        if isinstance(self.instance, Unset):
            instance = UNSET

        else:
            instance = self.instance

        login_token: Union[None, Unset, str]
        if isinstance(self.login_token, Unset):
            login_token = UNSET

        else:
            login_token = self.login_token

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "status": status,
                "type": type,
                "title": title,
                "detail": detail,
                "challenge": challenge,
            }
        )
        if instance is not UNSET:
            field_dict["instance"] = instance
        if login_token is not UNSET:
            field_dict["login_token"] = login_token

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        status = d.pop("status")

        type = d.pop("type")

        title = d.pop("title")

        detail = d.pop("detail")

        challenge = ChallengeResponseChallenge(d.pop("challenge"))

        def _parse_instance(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        instance = _parse_instance(d.pop("instance", UNSET))

        def _parse_login_token(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        login_token = _parse_login_token(d.pop("login_token", UNSET))

        challenge_response = cls(
            status=status,
            type=type,
            title=title,
            detail=detail,
            challenge=challenge,
            instance=instance,
            login_token=login_token,
        )

        challenge_response.additional_properties = d
        return challenge_response

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties





from dataclasses import dataclass
import io
import json
import pathlib
import ssl
from typing import IO, BinaryIO, Optional, TypeAlias, TypedDict, Union
import typing
import urllib.parse as urllib_parse
import pandas

import pyarrow.feather as feather

from invariant_client import pysdk
from invariant_client.bindings.invariant_instance_client.api.organization.ui_status_organization_name_api_v_1_ui_get import sync_detailed as ui_status_organization_name_api_v_1_ui_get

from invariant_client.bindings.invariant_instance_client.client import AuthenticatedClient as InstanceAuthenticatedClient
from invariant_client.bindings.invariant_instance_client import models
from invariant_client.bindings.invariant_instance_client import types
from invariant_client.bindings.invariant_instance_client.api.organization.list_reports_organization_name_api_v_1_reports_get import sync_detailed as list_reports_organization_name_api_v_1_reports_get
from invariant_client.bindings.invariant_instance_client.api.organization.refresh_organization_name_api_v1_refresh_post import sync_detailed as refresh_organization_name_api_v1_refresh_post
from invariant_client.bindings.invariant_instance_client.models.report_text_summary_request import ReportTextSummaryRequest
from invariant_client.bindings.invariant_login_client.client import AuthenticatedClient as LoginAuthenticatedClient, Client as LoginClient
from invariant_client.bindings.invariant_instance_client.api.organization.upload_snapshot_organization_name_api_v_1_uploadsnapshot_post import sync_detailed as upload_snapshot_organization_name_api_v_1_uploadsnapshot_post
from invariant_client.bindings.invariant_instance_client.api.organization.upload_snapshot_status_organization_name_api_v_1_uploadsnapshot_status_get import sync_detailed as upload_snapshot_status_organization_name_api_v_1_uploadsnapshot_status_get
from invariant_client.bindings.invariant_instance_client.api.organization.get_report_summary_organization_name_api_v_1_reports_report_id_summary_get import sync_detailed as get_report_summary_organization_name_api_v_1_reports_report_id_summary_get
from invariant_client.bindings.invariant_instance_client.api.organization.get_report_organization_name_api_v_1_reports_report_id_get import _get_kwargs as get_report_organization_name_api_v_1_reports_report_id_get__get_kwargs
from invariant_client.bindings.invariant_instance_client.api.organization.get_report_text_summary_organization_name_api_v_1_reports_report_id_text_get import sync_detailed as get_report_text_summary_organization_name_api_v_1_reports_report_id_text_get

from invariant_client.bindings.invariant_login_client.api.login.get_instances_api_v1_login_get_instances_post import sync_detailed as get_instances_api_v1_login_get_instances_post
from invariant_client.bindings.invariant_login_client.models.base_error_response import BaseErrorResponse
from invariant_client.bindings.invariant_login_client.models.validation_error_response import ValidationErrorResponse


DOMAIN_NAME = "https://invariant.tech"


class NoOrganization(Exception):
    """Credentials must be paired with an organization name."""


ErrorResponseType: TypeAlias = BaseErrorResponse | ValidationErrorResponse


class RemoteError(Exception):
    """Generic server-side or connection error."""


class AuthorizationException(Exception):
    """Server authentication rejected."""


class Settings(TypedDict):
    debug: bool


@dataclass
class AccessCredential:
    """An invariant access credential."""

    access_token: Optional[str]
    refresh_token: Optional[str]
    organization_name: str

    @classmethod
    def from_env(cls, env_data: dict[str, str], base_url: Optional[str] = None, verify_ssl: Optional[str | bool | ssl.SSLContext] = None) -> Optional['AccessCredential']:
        access_token = env_data.get(cls.INVARIANT_ACCESS_TOKEN)
        refresh_token = env_data.get(cls.INVARIANT_API_TOKEN)
        organization_name = env_data.get(cls.INVARIANT_ORGANIZATION_NAME)
        return cls.build(organization_name, access_token, refresh_token, base_url=base_url, verify_ssl=verify_ssl)

    @classmethod
    def from_file(cls, file_path: 'typing.Union[pathlib.Path, str]', base_url: Optional[str] = None, verify_ssl: Optional[str | bool | ssl.SSLContext] = None) -> Optional['AccessCredential']:
        with open(file_path, "r") as f:
            data: dict = json.load(f)
        access_token = data.get(cls.INVARIANT_ACCESS_TOKEN)
        refresh_token = data.get(cls.INVARIANT_API_TOKEN)
        organization_name = data.get(cls.INVARIANT_ORGANIZATION_NAME)
        return cls.build(organization_name, access_token, refresh_token, base_url=base_url, verify_ssl=verify_ssl)

    @classmethod
    def build(cls, organization_name: str, access_token: Optional[str] = None, refresh_token: Optional[str] = None, base_url: Optional[str] = None, verify_ssl: Optional[str | bool | ssl.SSLContext] = None):
        # if access_token, try it (get ui? org?)
        # - if error and no RT, error to user
        # if refresh_token, but no organization_name, error to user
        # if refresh_token, try it, and error to user if error
        # otherwise just return None
        if not (organization_name or access_token or refresh_token):
            return None

        error: Exception = None
        verify_ssl = verify_ssl or ssl.create_default_context()
        if not organization_name:
            raise NoOrganization("INVARIANT_ORGANIZATION_NAME must be given.")
        if access_token:
            creds = cls(access_token=access_token, refresh_token=None, organization_name=organization_name)
            client = pysdk.Invariant(creds=creds, settings={}, base_url=base_url, verify_ssl=verify_ssl)
            try:
                client.status() # Will throw if access token is no good
                return creds
            except RemoteError as r_error:
                error = r_error
            # try it
            # try to set organization_name
            # if error and no RT, error to user
            # if error and RT, pass thru to RT
        if refresh_token:
            creds = cls(refresh_token=refresh_token, organization_name=organization_name, access_token=None)
            base_url = base_url or DOMAIN_NAME
            client = pysdk.InvariantLogin(settings={}, creds=creds, base_url=base_url, verify_ssl=verify_ssl)
            try:
                client.get_instances() # Will throw if refresh token is no good
                sdk = client.to_instance_sdk(verify_ssl)
                return sdk.creds
            except RemoteError as r_error:
                error = r_error

        if error:
            raise error

    def to_json(self) -> str:
        data = {}
        if self.access_token:
            data[AccessCredential.INVARIANT_ACCESS_TOKEN] = self.access_token
        if self.refresh_token:
            data[AccessCredential.INVARIANT_API_TOKEN] = self.refresh_token
        if self.organization_name:
            data[AccessCredential.INVARIANT_ORGANIZATION_NAME] = self.organization_name
        return json.dumps(data)


AccessCredential.INVARIANT_ACCESS_TOKEN = 'INVARIANT_ACCESS_TOKEN'
AccessCredential.INVARIANT_API_TOKEN = 'INVARIANT_API_TOKEN'
AccessCredential.INVARIANT_ORGANIZATION_NAME = 'INVARIANT_ORGANIZATION_NAME'


class Invariant:

    client: InstanceAuthenticatedClient
    creds: AccessCredential
    base_url: str

    def __init__(
            self,
            creds: AccessCredential,
            settings: dict,
            base_url: Optional[str] = None,
            verify_ssl: Optional[str | bool | ssl.SSLContext] = None,
            **kwargs):
        self.creds = creds
        self.settings = settings
        base_url = base_url or DOMAIN_NAME
        self.base_url = self.app_base_url(base_url)

        # Prefer to use the Python default SSL context over the HTTPX SSL context, which does not consider system trust roots
        # Users can revert to the HTTPX SSL context with 'verify_ssl=True'
        verify_ssl = verify_ssl or ssl.create_default_context()
        self.client = InstanceAuthenticatedClient(
            self.base_url,
            token=creds.access_token,
            verify_ssl=verify_ssl,
            **kwargs)
    
    @staticmethod
    def app_base_url(base_domain_name: str) -> str:
        url = urllib_parse.urlparse(base_domain_name)
        url = url._replace(netloc=f'app.{url.netloc}')
        return url.geturl()
    
    def upload_snapshot(
            self,
            source: 'Union[IO, BinaryIO]',
            compare_to: Optional[str] = None) -> models.UploadSnapshotResponse:
        """Zip and upload the current folder. Display a summary of processing results when complete."""
        response = upload_snapshot_organization_name_api_v_1_uploadsnapshot_post(
            self.creds.organization_name,
            client=self.client,
            multipart_data=models.BodyUploadSnapshotOrganizationNameApiV1UploadsnapshotPost(
                file=types.File(
                    payload=source,
                    file_name="snapshot_upload.zip",
                    mime_type="application/zip"
                )
            ))
        response = response.parsed
        # TODO idiom should be to examine error responses by status code as we do in the UI
        # Note that DNS NX_DOMAIN error actually raises httpx.ConnectError (-2: Name or service not known) at request time
        # Connection refused also raises httpx.ConnectError (111: Connection refused)
        if not response:
            raise RemoteError(f"Unable to connect to {self.base_url}")
        if isinstance(response, models.ChallengeResponse):
            raise AuthorizationException(f"{response.title}: {response.detail}")
        if not isinstance(response, models.UploadSnapshotResponse):
            raise RemoteError(response)
        return response
    
    def upload_is_running(self, uuid: str) -> models.UploadSnapshotStatusResponse:
        response = upload_snapshot_status_organization_name_api_v_1_uploadsnapshot_status_get(
            self.creds.organization_name,
            client=self.client,
            uuid=uuid,
        )
        response = response.parsed
        if not response:
            raise RemoteError(f"Unable to connect to {self.base_url}")
        if isinstance(response, models.ChallengeResponse):
            raise AuthorizationException(f"{response.title}: {response.detail}")
        if not isinstance(response, models.UploadSnapshotStatusResponse):
            raise RemoteError(response)
        return response

    def list_snapshots(self):
        response = list_reports_organization_name_api_v_1_reports_get(self.creds.organization_name, client=self.client)
        response = response.parsed
        if not response:
            raise RemoteError(f"Unable to connect to {self.base_url}")
        if isinstance(response, models.ChallengeResponse):
            raise AuthorizationException(f"{response.title}: {response.detail}")
        if not isinstance(response, models.ListReportsResponse):
            raise RemoteError(response)
        return response

    def snapshot_detail(
            self,
            report_uuid: str):
        response = get_report_summary_organization_name_api_v_1_reports_report_id_summary_get(self.creds.organization_name, report_uuid, client=self.client)
        response = response.parsed
        if not response:
            raise RemoteError(f"Unable to connect to {self.base_url}")
        if isinstance(response, models.ChallengeResponse):
            raise AuthorizationException(f"{response.title}: {response.detail}")
        if not isinstance(response, models.GetReportSummaryResponse):
            raise RemoteError(response)
        return response

    def snapshot_file(
            self,
            file_uuid: str) -> pandas.DataFrame:
        kwargs = get_report_organization_name_api_v_1_reports_report_id_get__get_kwargs(
            organization_name=self.creds.organization_name,
            report_id=file_uuid,
        )
        response = self.client.get_httpx_client().request(
            **kwargs,
        )
        
        # TODO approach checking for errors more carefully as the expected value is not JSON
        if not response:
            raise RemoteError(f"Unable to connect to {self.base_url}")
        # if isinstance(response, models.ChallengeResponse):
        #     raise AuthorizationException(f"{response.title}: {response.detail}")
        # if isinstance(response, models.BaseErrorResponse):
        #     raise RemoteError(response)
        response: pandas.DataFrame = feather.read_feather(io.BytesIO(response.content))
        return response

    def snapshot_file_text(
            self,
            file_uuid: str,
            traces: bool,
            json_mode: bool):
        response = get_report_text_summary_organization_name_api_v_1_reports_report_id_text_get(
            self.creds.organization_name,
            file_uuid,
            client=self.client,
            json_body=ReportTextSummaryRequest(traces=traces, mode='json' if json_mode else 'text'))
        response = response.parsed
        if not response:
            raise RemoteError(f"Unable to connect to {self.base_url}")
        if isinstance(response, models.ChallengeResponse):
            raise AuthorizationException(f"{response.title}: {response.detail}")
        if not isinstance(response, models.ReportTextSummaryResponse):
            raise RemoteError(response)
        return response

    def show(
            self,
            snapshot: str,
            file: str):
        pass

    def show_solution(
            self,
            snapshot: str,
            solution: str):
        pass

    def status(self) -> models.UIStatusResponse:
        if not self.creds or not self.creds.access_token:
            raise ValueError("status requires an access token.")
        response = ui_status_organization_name_api_v_1_ui_get(self.creds.organization_name, client=self.client)
        response = response.parsed
        if not response:
            raise RemoteError(f"Unable to connect to {self.base_url}")
        if isinstance(response, models.ChallengeResponse):
            raise AuthorizationException(f"{response.title}: {response.detail}")
        if not isinstance(response, models.UIStatusResponse):
            raise RemoteError(response)
        return response


class InvariantLogin:

    client: LoginAuthenticatedClient
    login_client: LoginAuthenticatedClient
    creds: AccessCredential
    login_creds: AccessCredential
    base_url: str

    def __init__(
            self,
            settings: dict,
            creds: Optional[AccessCredential] = None,
            login_session_creds: Optional[AccessCredential] = None,
            base_url: Optional[str] = None,
            verify_ssl: Optional[str | bool | ssl.SSLContext] = None,
            **kwargs):
        self.creds = creds
        self.login_creds = login_session_creds
        self.settings = settings
        self.base_url = base_url or DOMAIN_NAME

        # Prefer to use the Python default SSL context over the HTTPX SSL context, which does not consider system trust roots
        # Users can revert to the HTTPX SSL context with 'verify_ssl=True'
        verify_ssl = verify_ssl or ssl.create_default_context()

        # Three credential modes for the login service: no creds, login session, and refresh token
        if creds.refresh_token:
            self.client = LoginClient(
                self.base_url,
                cookies={'refresh_token_cookie': creds.refresh_token},
                verify_ssl=verify_ssl,
                **kwargs)
        elif login_session_creds:
            self.login_client = LoginAuthenticatedClient(
                self.base_url,
                token=login_session_creds.access_token,
                verify_ssl=verify_ssl,
                **kwargs)
        else:
            self.login_client = LoginClient(
                self.base_url,
                verify_ssl=verify_ssl,
                **kwargs)

    def get_instances(self):
        if not self.creds or not self.creds.refresh_token:
            raise ValueError("get_instances requires a refresh token.")
        response = get_instances_api_v1_login_get_instances_post(client=self.client)
        response = response.parsed
        if not response:
            raise RemoteError(f"Unable to connect to {self.base_url}")
        if isinstance(response, models.ChallengeResponse):
            raise AuthorizationException(f"{response.title}: {response.detail}")
        if not isinstance(response, list):
            raise RemoteError(response)
        return response

    def to_instance_sdk(self, verify_ssl: str | bool | ssl.SSLContext = True):
        if not self.creds or not self.creds.refresh_token or not self.creds.organization_name:
            raise ValueError("to_instance_sdk requires a refresh token and organization name.")
        base_url = Invariant.app_base_url(self.base_url)
        client = LoginClient(
            base_url,
            cookies={'refresh_token_cookie': self.creds.refresh_token},
            verify_ssl=verify_ssl)

        response = refresh_organization_name_api_v1_refresh_post(organization_name=self.creds.organization_name, client=client)
        response = response.parsed
        if not response:
            raise RemoteError(f"Unable to connect to {self.base_url}")
        if isinstance(response, models.ChallengeResponse):
            raise AuthorizationException(f"{response.title}: {response.detail}")
        if not isinstance(response, models.RefreshResponse):
            raise RemoteError(response)

        new_creds = AccessCredential(
            response.access_token,
            refresh_token=self.creds.refresh_token,
            organization_name=self.creds.organization_name)
        return Invariant(
            creds=new_creds,
            settings=self.settings,
            base_url=self.base_url,
            verify_ssl=verify_ssl)



from typing import Any, Dict, Type, TypeVar

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field
import json

from ..types import UNSET

from typing import Union
from typing import List
from ..types import File, FileJsonType
from io import BytesIO


T = TypeVar("T", bound="BodyUploadSnapshotOrganizationNameApiV1UploadsnapshotPost")


@_attrs_define
class BodyUploadSnapshotOrganizationNameApiV1UploadsnapshotPost:
    """
    Attributes:
        file (Union[File, List[File]]):
    """

    file: Union[File, List[File]]
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        file: Union[FileJsonType, List[FileJsonType]]

        if isinstance(self.file, File):
            file = self.file.to_tuple()

        else:
            file = []
            for file_type_1_item_data in self.file:
                file_type_1_item = file_type_1_item_data.to_tuple()

                file.append(file_type_1_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "file": file,
            }
        )

        return field_dict

    def to_multipart(self) -> Dict[str, Any]:
        file: Union[File, List[File]]

        if isinstance(self.file, File):
            file = self.file.to_tuple()

        else:
            _temp_file = []
            for file_type_1_item_data in self.file:
                file_type_1_item = file_type_1_item_data.to_tuple()

                _temp_file.append(file_type_1_item)
            file = (None, json.dumps(_temp_file).encode(), "application/json")

        field_dict: Dict[str, Any] = {}
        field_dict.update(
            {
                key: (None, str(value).encode(), "text/plain")
                for key, value in self.additional_properties.items()
            }
        )
        field_dict.update(
            {
                "file": file,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()

        def _parse_file(data: object) -> Union[File, List[File]]:
            try:
                if not isinstance(data, bytes):
                    raise TypeError()
                file_type_0 = File(payload=BytesIO(data))

                return file_type_0
            except:  # noqa: E722
                pass
            if not isinstance(data, list):
                raise TypeError()
            file_type_1 = UNSET
            _file_type_1 = data
            for file_type_1_item_data in _file_type_1:
                file_type_1_item = File(payload=BytesIO(file_type_1_item_data))

                file_type_1.append(file_type_1_item)

            return file_type_1

        file = _parse_file(d.pop("file"))

        body_upload_snapshot_organization_name_api_v1_uploadsnapshot_post = cls(
            file=file,
        )

        body_upload_snapshot_organization_name_api_v1_uploadsnapshot_post.additional_properties = d
        return body_upload_snapshot_organization_name_api_v1_uploadsnapshot_post

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, Type, TypeVar, TYPE_CHECKING

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field


from typing import cast, Union
from typing import cast
from typing import Dict

if TYPE_CHECKING:
    from ..models.tab_info_state_type_0 import TabInfoStateType0
    from ..models.tab_info_parameters_type_0 import TabInfoParametersType0


T = TypeVar("T", bound="TabInfo")


@_attrs_define
class TabInfo:
    """
    Attributes:
        urn (str):
        active (bool):
        override_name (Union[None, str]):
        parameters (Union['TabInfoParametersType0', None]):
        state (Union['TabInfoStateType0', None]):
    """

    urn: str
    active: bool
    override_name: Union[None, str]
    parameters: Union["TabInfoParametersType0", None]
    state: Union["TabInfoStateType0", None]
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        from ..models.tab_info_state_type_0 import TabInfoStateType0
        from ..models.tab_info_parameters_type_0 import TabInfoParametersType0

        urn = self.urn
        active = self.active
        override_name: Union[None, str]

        override_name = self.override_name

        parameters: Union[Dict[str, Any], None]

        if isinstance(self.parameters, TabInfoParametersType0):
            parameters = self.parameters.to_dict()

        else:
            parameters = self.parameters

        state: Union[Dict[str, Any], None]

        if isinstance(self.state, TabInfoStateType0):
            state = self.state.to_dict()

        else:
            state = self.state

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "urn": urn,
                "active": active,
                "override_name": override_name,
                "parameters": parameters,
                "state": state,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.tab_info_state_type_0 import TabInfoStateType0
        from ..models.tab_info_parameters_type_0 import TabInfoParametersType0

        d = src_dict.copy()
        urn = d.pop("urn")

        active = d.pop("active")

        def _parse_override_name(data: object) -> Union[None, str]:
            if data is None:
                return data
            return cast(Union[None, str], data)

        override_name = _parse_override_name(d.pop("override_name"))

        def _parse_parameters(data: object) -> Union["TabInfoParametersType0", None]:
            if data is None:
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                parameters_type_0 = TabInfoParametersType0.from_dict(data)

                return parameters_type_0
            except:  # noqa: E722
                pass
            return cast(Union["TabInfoParametersType0", None], data)

        parameters = _parse_parameters(d.pop("parameters"))

        def _parse_state(data: object) -> Union["TabInfoStateType0", None]:
            if data is None:
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                state_type_0 = TabInfoStateType0.from_dict(data)

                return state_type_0
            except:  # noqa: E722
                pass
            return cast(Union["TabInfoStateType0", None], data)

        state = _parse_state(d.pop("state"))

        tab_info = cls(
            urn=urn,
            active=active,
            override_name=override_name,
            parameters=parameters,
            state=state,
        )

        tab_info.additional_properties = d
        return tab_info

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, Type, TypeVar, TYPE_CHECKING

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

from typing import Union
from typing import cast, Union
from typing import cast
from ..types import UNSET, Unset
from typing import Dict

if TYPE_CHECKING:
    from ..models.comparison_reportdata_files import ComparisonReportdataFiles


T = TypeVar("T", bound="ComparisonReportdata")


@_attrs_define
class ComparisonReportdata:
    """
    Attributes:
        files (ComparisonReportdataFiles):
        solutions (Union[None, Unset, str]):
    """

    files: "ComparisonReportdataFiles"
    solutions: Union[None, Unset, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        files = self.files.to_dict()

        solutions: Union[None, Unset, str]
        if isinstance(self.solutions, Unset):
            solutions = UNSET

        else:
            solutions = self.solutions

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "files": files,
            }
        )
        if solutions is not UNSET:
            field_dict["solutions"] = solutions

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.comparison_reportdata_files import ComparisonReportdataFiles

        d = src_dict.copy()
        files = ComparisonReportdataFiles.from_dict(d.pop("files"))

        def _parse_solutions(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        solutions = _parse_solutions(d.pop("solutions", UNSET))

        comparison_reportdata = cls(
            files=files,
            solutions=solutions,
        )

        comparison_reportdata.additional_properties = d
        return comparison_reportdata

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from http import HTTPStatus
from typing import Any, Dict, Optional, Union, cast

import httpx

from ...client import AuthenticatedClient, Client
from ...types import Response
from ... import errors

from ...models.challenge_response import ChallengeResponse
from ...models.validation_error_response import ValidationErrorResponse
from typing import cast
from typing import Dict
from ...models.base_error_response import BaseErrorResponse


def _get_kwargs(
    organization_name: str,
    report_id: str,
) -> Dict[str, Any]:
    return {
        "method": "get",
        "url": "/{organization_name}/api/v1/reports/{report_id}/download".format(
            organization_name=organization_name,
            report_id=report_id,
        ),
    }


def _parse_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Optional[
    Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]
]:
    if response.status_code == HTTPStatus.OK:
        response_200 = cast(Any, None)
        return response_200
    if response.status_code == HTTPStatus.UNPROCESSABLE_ENTITY:
        response_422 = ValidationErrorResponse.from_dict(response.json())

        return response_422
    if response.status_code == HTTPStatus.UNAUTHORIZED:
        response_401 = ChallengeResponse.from_dict(response.json())

        return response_401
    if response.status_code == HTTPStatus.NOT_FOUND:
        response_404 = BaseErrorResponse.from_dict(response.json())

        return response_404
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[
    Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]
]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    organization_name: str,
    report_id: str,
    *,
    client: AuthenticatedClient,
) -> Response[
    Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]
]:
    """Retrieves a single report for an organization.

    Args:
        organization_name (str):
        report_id (str):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]]
    """

    kwargs = _get_kwargs(
        organization_name=organization_name,
        report_id=report_id,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    organization_name: str,
    report_id: str,
    *,
    client: AuthenticatedClient,
) -> Optional[
    Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]
]:
    """Retrieves a single report for an organization.

    Args:
        organization_name (str):
        report_id (str):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]
    """

    return sync_detailed(
        organization_name=organization_name,
        report_id=report_id,
        client=client,
    ).parsed


async def asyncio_detailed(
    organization_name: str,
    report_id: str,
    *,
    client: AuthenticatedClient,
) -> Response[
    Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]
]:
    """Retrieves a single report for an organization.

    Args:
        organization_name (str):
        report_id (str):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]]
    """

    kwargs = _get_kwargs(
        organization_name=organization_name,
        report_id=report_id,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    organization_name: str,
    report_id: str,
    *,
    client: AuthenticatedClient,
) -> Optional[
    Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]
]:
    """Retrieves a single report for an organization.

    Args:
        organization_name (str):
        report_id (str):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]
    """

    return (
        await asyncio_detailed(
            organization_name=organization_name,
            report_id=report_id,
            client=client,
        )
    ).parsed



from typing import Any, Dict, Type, TypeVar

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field


T = TypeVar("T", bound="CreateIntegrationRequestGithubAppInstallationData")


@_attrs_define
class CreateIntegrationRequestGithubAppInstallationData:
    """
    Attributes:
        code (str):
        installation_id (str):
        setup_action (str):
        state (str):
    """

    code: str
    installation_id: str
    setup_action: str
    state: str
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        code = self.code
        installation_id = self.installation_id
        setup_action = self.setup_action
        state = self.state

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "code": code,
                "installation_id": installation_id,
                "setup_action": setup_action,
                "state": state,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        code = d.pop("code")

        installation_id = d.pop("installation_id")

        setup_action = d.pop("setup_action")

        state = d.pop("state")

        create_integration_request_github_app_installation_data = cls(
            code=code,
            installation_id=installation_id,
            setup_action=setup_action,
            state=state,
        )

        create_integration_request_github_app_installation_data.additional_properties = d
        return create_integration_request_github_app_installation_data

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



""" Contains all the data models used in inputs/outputs """

from .base_error_response import BaseErrorResponse
from .challenge_response import ChallengeResponse
from .challenge_response_challenge import ChallengeResponseChallenge
from .check_invite_response import CheckInviteResponse
from .consume_client_login_session_response import ConsumeClientLoginSessionResponse
from .create_client_login_session_response import CreateClientLoginSessionResponse
from .email_check_request import EmailCheckRequest
from .email_password_login import EmailPasswordLogin
from .fulfill_client_login_request import FulfillClientLoginRequest
from .login import Login
from .login_request import LoginRequest
from .new_password_request import NewPasswordRequest
from .new_password_request_authn_type import NewPasswordRequestAuthnType
from .organization import Organization
from .register_organization_request_body import RegisterOrganizationRequestBody
from .reset_pin_request import ResetPINRequest
from .reset_request import ResetRequest
from .user import User
from .validation_error_response import ValidationErrorResponse
from .validation_error_response_part import ValidationErrorResponsePart
from .validation_request import ValidationRequest

__all__ = (
    "BaseErrorResponse",
    "ChallengeResponse",
    "ChallengeResponseChallenge",
    "CheckInviteResponse",
    "ConsumeClientLoginSessionResponse",
    "CreateClientLoginSessionResponse",
    "EmailCheckRequest",
    "EmailPasswordLogin",
    "FulfillClientLoginRequest",
    "Login",
    "LoginRequest",
    "NewPasswordRequest",
    "NewPasswordRequestAuthnType",
    "Organization",
    "RegisterOrganizationRequestBody",
    "ResetPINRequest",
    "ResetRequest",
    "User",
    "ValidationErrorResponse",
    "ValidationErrorResponsePart",
    "ValidationRequest",
)



from typing import Any, Dict, Type, TypeVar, TYPE_CHECKING

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field


from typing import List
from typing import Dict

if TYPE_CHECKING:
    from ..models.tab_info import TabInfo


T = TypeVar("T", bound="UserTabsConfig")


@_attrs_define
class UserTabsConfig:
    """
    Attributes:
        tabs (List['TabInfo']):
    """

    tabs: List["TabInfo"]
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        tabs = []
        for tabs_item_data in self.tabs:
            tabs_item = tabs_item_data.to_dict()

            tabs.append(tabs_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "tabs": tabs,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.tab_info import TabInfo

        d = src_dict.copy()
        tabs = []
        _tabs = d.pop("tabs")
        for tabs_item_data in _tabs:
            tabs_item = TabInfo.from_dict(tabs_item_data)

            tabs.append(tabs_item)

        user_tabs_config = cls(
            tabs=tabs,
        )

        user_tabs_config.additional_properties = d
        return user_tabs_config

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, Type, TypeVar

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field


from typing import cast, Union


T = TypeVar("T", bound="CreateTokenRequest")


@_attrs_define
class CreateTokenRequest:
    """
    Attributes:
        name (Union[None, str]):
        comment (Union[None, str]):
    """

    name: Union[None, str]
    comment: Union[None, str]
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        name: Union[None, str]

        name = self.name

        comment: Union[None, str]

        comment = self.comment

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "name": name,
                "comment": comment,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()

        def _parse_name(data: object) -> Union[None, str]:
            if data is None:
                return data
            return cast(Union[None, str], data)

        name = _parse_name(d.pop("name"))

        def _parse_comment(data: object) -> Union[None, str]:
            if data is None:
                return data
            return cast(Union[None, str], data)

        comment = _parse_comment(d.pop("comment"))

        create_token_request = cls(
            name=name,
            comment=comment,
        )

        create_token_request.additional_properties = d
        return create_token_request

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, Type, TypeVar, TYPE_CHECKING

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field


from typing import Dict

if TYPE_CHECKING:
    from ..models.login import Login
    from ..models.organization import Organization


T = TypeVar("T", bound="User")


@_attrs_define
class User:
    """
    Attributes:
        uuid (str):
        organization (Organization):
        login (Login):
        email (str):
        is_active (bool):
        is_superuser (bool):
    """

    uuid: str
    organization: "Organization"
    login: "Login"
    email: str
    is_active: bool
    is_superuser: bool
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        uuid = self.uuid
        organization = self.organization.to_dict()

        login = self.login.to_dict()

        email = self.email
        is_active = self.is_active
        is_superuser = self.is_superuser

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "uuid": uuid,
                "organization": organization,
                "login": login,
                "email": email,
                "is_active": is_active,
                "is_superuser": is_superuser,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.login import Login
        from ..models.organization import Organization

        d = src_dict.copy()
        uuid = d.pop("uuid")

        organization = Organization.from_dict(d.pop("organization"))

        login = Login.from_dict(d.pop("login"))

        email = d.pop("email")

        is_active = d.pop("is_active")

        is_superuser = d.pop("is_superuser")

        user = cls(
            uuid=uuid,
            organization=organization,
            login=login,
            email=email,
            is_active=is_active,
            is_superuser=is_superuser,
        )

        user.additional_properties = d
        return user

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, Type, TypeVar, TYPE_CHECKING

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

from typing import Union
from typing import cast, Union
from typing import cast
from ..types import UNSET, Unset
from dateutil.parser import isoparse
from typing import Dict
import datetime

if TYPE_CHECKING:
    from ..models.user_metadata import UserMetadata


T = TypeVar("T", bound="User")


@_attrs_define
class User:
    """
    Attributes:
        uuid (str):
        organization_uuid (str):
        login_uuid (str):
        email (str):
        metadata (UserMetadata):
        is_active (bool):
        created_at (datetime.datetime):
        deleted_at (Union[None, Unset, datetime.datetime]):
    """

    uuid: str
    organization_uuid: str
    login_uuid: str
    email: str
    metadata: "UserMetadata"
    is_active: bool
    created_at: datetime.datetime
    deleted_at: Union[None, Unset, datetime.datetime] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        uuid = self.uuid
        organization_uuid = self.organization_uuid
        login_uuid = self.login_uuid
        email = self.email
        metadata = self.metadata.to_dict()

        is_active = self.is_active
        created_at = self.created_at.isoformat()

        deleted_at: Union[None, Unset, str]
        if isinstance(self.deleted_at, Unset):
            deleted_at = UNSET

        elif isinstance(self.deleted_at, datetime.datetime):
            deleted_at = UNSET
            if not isinstance(self.deleted_at, Unset):
                deleted_at = self.deleted_at.isoformat()

        else:
            deleted_at = self.deleted_at

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "uuid": uuid,
                "organization_uuid": organization_uuid,
                "login_uuid": login_uuid,
                "email": email,
                "metadata": metadata,
                "is_active": is_active,
                "created_at": created_at,
            }
        )
        if deleted_at is not UNSET:
            field_dict["deleted_at"] = deleted_at

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.user_metadata import UserMetadata

        d = src_dict.copy()
        uuid = d.pop("uuid")

        organization_uuid = d.pop("organization_uuid")

        login_uuid = d.pop("login_uuid")

        email = d.pop("email")

        metadata = UserMetadata.from_dict(d.pop("metadata"))

        is_active = d.pop("is_active")

        created_at = isoparse(d.pop("created_at"))

        def _parse_deleted_at(data: object) -> Union[None, Unset, datetime.datetime]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, str):
                    raise TypeError()
                _deleted_at_type_0 = data
                deleted_at_type_0: Union[Unset, datetime.datetime]
                if isinstance(_deleted_at_type_0, Unset):
                    deleted_at_type_0 = UNSET
                else:
                    deleted_at_type_0 = isoparse(_deleted_at_type_0)

                return deleted_at_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, Unset, datetime.datetime], data)

        deleted_at = _parse_deleted_at(d.pop("deleted_at", UNSET))

        user = cls(
            uuid=uuid,
            organization_uuid=organization_uuid,
            login_uuid=login_uuid,
            email=email,
            metadata=metadata,
            is_active=is_active,
            created_at=created_at,
            deleted_at=deleted_at,
        )

        user.additional_properties = d
        return user

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from http import HTTPStatus
from typing import Any, Dict, Optional, Union, cast

import httpx

from ...client import AuthenticatedClient, Client
from ...types import Response
from ... import errors

from typing import cast
from typing import Dict
from ...models.challenge_response import ChallengeResponse
from ...models.validation_error_response import ValidationErrorResponse


def _get_kwargs() -> Dict[str, Any]:
    return {
        "method": "post",
        "url": "/api/v1/login/refresh",
    }


def _parse_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Optional[Union[Any, ChallengeResponse, ValidationErrorResponse]]:
    if response.status_code == HTTPStatus.NO_CONTENT:
        response_204 = cast(Any, None)
        return response_204
    if response.status_code == HTTPStatus.UNPROCESSABLE_ENTITY:
        response_422 = ValidationErrorResponse.from_dict(response.json())

        return response_422
    if response.status_code == HTTPStatus.UNAUTHORIZED:
        response_401 = ChallengeResponse.from_dict(response.json())

        return response_401
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[Union[Any, ChallengeResponse, ValidationErrorResponse]]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    *,
    client: AuthenticatedClient,
) -> Response[Union[Any, ChallengeResponse, ValidationErrorResponse]]:
    """Issue a refresh token from a valid, non-expired refresh token

     Issue a refresh token from a valid, non-expired refresh token.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Any, ChallengeResponse, ValidationErrorResponse]]
    """

    kwargs = _get_kwargs()

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    *,
    client: AuthenticatedClient,
) -> Optional[Union[Any, ChallengeResponse, ValidationErrorResponse]]:
    """Issue a refresh token from a valid, non-expired refresh token

     Issue a refresh token from a valid, non-expired refresh token.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Any, ChallengeResponse, ValidationErrorResponse]
    """

    return sync_detailed(
        client=client,
    ).parsed


async def asyncio_detailed(
    *,
    client: AuthenticatedClient,
) -> Response[Union[Any, ChallengeResponse, ValidationErrorResponse]]:
    """Issue a refresh token from a valid, non-expired refresh token

     Issue a refresh token from a valid, non-expired refresh token.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Any, ChallengeResponse, ValidationErrorResponse]]
    """

    kwargs = _get_kwargs()

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    *,
    client: AuthenticatedClient,
) -> Optional[Union[Any, ChallengeResponse, ValidationErrorResponse]]:
    """Issue a refresh token from a valid, non-expired refresh token

     Issue a refresh token from a valid, non-expired refresh token.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Any, ChallengeResponse, ValidationErrorResponse]
    """

    return (
        await asyncio_detailed(
            client=client,
        )
    ).parsed



from typing import Any, Dict, Type, TypeVar, TYPE_CHECKING

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field


from typing import Literal
from typing import Dict

if TYPE_CHECKING:
    from ..models.integration_data_github_app_installation_data import (
        IntegrationDataGithubAppInstallationData,
    )


T = TypeVar("T", bound="IntegrationDataGithubAppInstallation")


@_attrs_define
class IntegrationDataGithubAppInstallation:
    """
    Attributes:
        type (Literal['github_app_installation']):
        github_app_install (IntegrationDataGithubAppInstallationData):
    """

    type: Literal["github_app_installation"]
    github_app_install: "IntegrationDataGithubAppInstallationData"
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        type = self.type
        github_app_install = self.github_app_install.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "type": type,
                "github_app_install": github_app_install,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.integration_data_github_app_installation_data import (
            IntegrationDataGithubAppInstallationData,
        )

        d = src_dict.copy()
        type = d.pop("type")

        github_app_install = IntegrationDataGithubAppInstallationData.from_dict(
            d.pop("github_app_install")
        )

        integration_data_github_app_installation = cls(
            type=type,
            github_app_install=github_app_install,
        )

        integration_data_github_app_installation.additional_properties = d
        return integration_data_github_app_installation

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, Type, TypeVar, TYPE_CHECKING

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field


from typing import Dict

if TYPE_CHECKING:
    from ..models.get_report_summary_response_summary import (
        GetReportSummaryResponseSummary,
    )
    from ..models.get_report_summary_response_status import (
        GetReportSummaryResponseStatus,
    )
    from ..models.report import Report


T = TypeVar("T", bound="GetReportSummaryResponse")


@_attrs_define
class GetReportSummaryResponse:
    """
    Attributes:
        report (Report):
        summary (GetReportSummaryResponseSummary):
        status (GetReportSummaryResponseStatus):
    """

    report: "Report"
    summary: "GetReportSummaryResponseSummary"
    status: "GetReportSummaryResponseStatus"
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        report = self.report.to_dict()

        summary = self.summary.to_dict()

        status = self.status.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "report": report,
                "summary": summary,
                "status": status,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.get_report_summary_response_summary import (
            GetReportSummaryResponseSummary,
        )
        from ..models.get_report_summary_response_status import (
            GetReportSummaryResponseStatus,
        )
        from ..models.report import Report

        d = src_dict.copy()
        report = Report.from_dict(d.pop("report"))

        summary = GetReportSummaryResponseSummary.from_dict(d.pop("summary"))

        status = GetReportSummaryResponseStatus.from_dict(d.pop("status"))

        get_report_summary_response = cls(
            report=report,
            summary=summary,
            status=status,
        )

        get_report_summary_response.additional_properties = d
        return get_report_summary_response

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, Type, TypeVar

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

from ..types import UNSET, Unset
from typing import Literal, Union


T = TypeVar("T", bound="CreateClientLoginSessionResponse")


@_attrs_define
class CreateClientLoginSessionResponse:
    """
    Attributes:
        status (int):
        pin (str):
        url (str):
        uuid (str):
        token (str):
        type (Union[Literal['urn:invariant:responses:init_client_login_response'], Unset]):  Default:
            'urn:invariant:responses:init_client_login_response'.
    """

    status: int
    pin: str
    url: str
    uuid: str
    token: str
    type: Union[
        Literal["urn:invariant:responses:init_client_login_response"], Unset
    ] = "urn:invariant:responses:init_client_login_response"
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        status = self.status
        pin = self.pin
        url = self.url
        uuid = self.uuid
        token = self.token
        type = self.type

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "status": status,
                "pin": pin,
                "url": url,
                "uuid": uuid,
                "token": token,
            }
        )
        if type is not UNSET:
            field_dict["type"] = type

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        status = d.pop("status")

        pin = d.pop("pin")

        url = d.pop("url")

        uuid = d.pop("uuid")

        token = d.pop("token")

        type = d.pop("type", UNSET)

        create_client_login_session_response = cls(
            status=status,
            pin=pin,
            url=url,
            uuid=uuid,
            token=token,
            type=type,
        )

        create_client_login_session_response.additional_properties = d
        return create_client_login_session_response

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import enum
import json
from attr import asdict
import pandas
from tabulate import tabulate
from invariant_client.bindings.invariant_instance_client.models.get_report_summary_response import GetReportSummaryResponse


class OutputFormat(enum.Enum):
    TABULATE = enum.auto()
    JSON = enum.auto()
    TSV = enum.auto()
    # etc


def snapshot_status(response: GetReportSummaryResponse):
    """Display a message indicating the overall status of the analysis user task."""
    status = response.status.to_dict()
    status = status.get('state')
    if status == 'COMPLETE':
        pass
    elif status == 'FAILED':
        print("Error: Snapshot could not be evaluated.")
    elif status == 'INCOMPLETE':
        print("Error: Snapshot evaluation did not finish.")


def snapshot_summary_table(response: GetReportSummaryResponse, format: OutputFormat):
    """Display a table containing row counts for all emitted reports."""
    if format == OutputFormat.JSON:
        pass
    else:
        print_frame(pandas.DataFrame(response.summary.to_dict().items(), columns=['File', 'RowCount']), format)


def snapshot_halted(response: GetReportSummaryResponse):
    """Describe each halted step."""
    status = response.status.to_dict()
    halted = []
    def visit_step(step: dict, prefix: list):
        prefix = prefix + [step['name']] if prefix else [step['name']]
        if step['state'] != 'COMPLETE':
            halted.append((prefix, step))
        for child_step in step['steps']:
            visit_step(child_step, prefix)
    visit_step(status, None)
    if len(halted):
        print("\nThe following steps were not completed:")
        for prefix, step in halted:
            print(f"    {' > '.join(prefix)}")
            if step['state'] != 'FAILED':
                print(f"\n        {step['state']}")


def snapshot_errors(errors: pandas.DataFrame, format: OutputFormat):
    """Describe each error."""
    # Group errors by label, then repeat the header every 10th member
    for label, error_group in errors.groupby(by=['label']):
        label = label[0]
        for i, error in enumerate(error_group.to_dict(orient='records')):
            prefix = ""
            if not i % 10:
                prefix = f"In {label}{' (continued)' if i > 0 else ''}:\n\n"
            data = json.loads(error['detail'])
            if data['type'] == 'urn:invariant:errors:child_step_failed':
                continue
            if data['type'] == 'urn:invariant:errors:internal_exception':
                print(f"\nInternal error in {label}:\n    {data['detail']}")
                continue

            print(f"\n{prefix}    {data['title']}\n    {data['detail']}")


def print_frame(data: pandas.DataFrame, format: OutputFormat):
    if format == OutputFormat.TSV:
        print(tabulate(data, headers='keys', tablefmt='tsv'))
    elif format == OutputFormat.TABULATE:
        print(tabulate(data, headers='keys', tablefmt='psql'))
    else:
        raise ValueError(f"Unacceptable format: {format}")



from http import HTTPStatus
from typing import Any, Dict, Optional, Union

import httpx

from ...client import AuthenticatedClient, Client
from ...types import Response
from ... import errors

from ...models.check_invite_response import CheckInviteResponse
from ...models.validation_error_response import ValidationErrorResponse
from ...models.challenge_response import ChallengeResponse
from ...models.base_error_response import BaseErrorResponse
from typing import Dict


def _get_kwargs() -> Dict[str, Any]:
    return {
        "method": "post",
        "url": "/api/v1/login/check_invite",
    }


def _parse_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Optional[
    Union[
        BaseErrorResponse,
        ChallengeResponse,
        CheckInviteResponse,
        ValidationErrorResponse,
    ]
]:
    if response.status_code == HTTPStatus.OK:
        response_200 = CheckInviteResponse.from_dict(response.json())

        return response_200
    if response.status_code == HTTPStatus.UNPROCESSABLE_ENTITY:
        response_422 = ValidationErrorResponse.from_dict(response.json())

        return response_422
    if response.status_code == HTTPStatus.UNAUTHORIZED:
        response_401 = ChallengeResponse.from_dict(response.json())

        return response_401
    if response.status_code == HTTPStatus.FORBIDDEN:
        response_403 = BaseErrorResponse.from_dict(response.json())

        return response_403
    if response.status_code == HTTPStatus.NOT_FOUND:
        response_404 = BaseErrorResponse.from_dict(response.json())

        return response_404
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[
    Union[
        BaseErrorResponse,
        ChallengeResponse,
        CheckInviteResponse,
        ValidationErrorResponse,
    ]
]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    *,
    client: AuthenticatedClient,
) -> Response[
    Union[
        BaseErrorResponse,
        ChallengeResponse,
        CheckInviteResponse,
        ValidationErrorResponse,
    ]
]:
    """Consume an email invitation magic link

     Consume an email invitation magic link.

    If the email is already associated with an active login, this will
    link the new user and active login and issue an HTTPOnly refresh token cookie.

    Otherwise, a ChallengeResponse will be issued with next=new_password.

    This function can also handle the case where the login was created with a pre-set password,
    although creating such a login may not be supported.

    This function also marks the linked login as email_validated=True.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[BaseErrorResponse, ChallengeResponse, CheckInviteResponse, ValidationErrorResponse]]
    """

    kwargs = _get_kwargs()

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    *,
    client: AuthenticatedClient,
) -> Optional[
    Union[
        BaseErrorResponse,
        ChallengeResponse,
        CheckInviteResponse,
        ValidationErrorResponse,
    ]
]:
    """Consume an email invitation magic link

     Consume an email invitation magic link.

    If the email is already associated with an active login, this will
    link the new user and active login and issue an HTTPOnly refresh token cookie.

    Otherwise, a ChallengeResponse will be issued with next=new_password.

    This function can also handle the case where the login was created with a pre-set password,
    although creating such a login may not be supported.

    This function also marks the linked login as email_validated=True.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[BaseErrorResponse, ChallengeResponse, CheckInviteResponse, ValidationErrorResponse]
    """

    return sync_detailed(
        client=client,
    ).parsed


async def asyncio_detailed(
    *,
    client: AuthenticatedClient,
) -> Response[
    Union[
        BaseErrorResponse,
        ChallengeResponse,
        CheckInviteResponse,
        ValidationErrorResponse,
    ]
]:
    """Consume an email invitation magic link

     Consume an email invitation magic link.

    If the email is already associated with an active login, this will
    link the new user and active login and issue an HTTPOnly refresh token cookie.

    Otherwise, a ChallengeResponse will be issued with next=new_password.

    This function can also handle the case where the login was created with a pre-set password,
    although creating such a login may not be supported.

    This function also marks the linked login as email_validated=True.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[BaseErrorResponse, ChallengeResponse, CheckInviteResponse, ValidationErrorResponse]]
    """

    kwargs = _get_kwargs()

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    *,
    client: AuthenticatedClient,
) -> Optional[
    Union[
        BaseErrorResponse,
        ChallengeResponse,
        CheckInviteResponse,
        ValidationErrorResponse,
    ]
]:
    """Consume an email invitation magic link

     Consume an email invitation magic link.

    If the email is already associated with an active login, this will
    link the new user and active login and issue an HTTPOnly refresh token cookie.

    Otherwise, a ChallengeResponse will be issued with next=new_password.

    This function can also handle the case where the login was created with a pre-set password,
    although creating such a login may not be supported.

    This function also marks the linked login as email_validated=True.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[BaseErrorResponse, ChallengeResponse, CheckInviteResponse, ValidationErrorResponse]
    """

    return (
        await asyncio_detailed(
            client=client,
        )
    ).parsed



#   Copyright 2018 The Batfish Open Source Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

# Modified 2024 Invariant Technology, Inc
# The following modifications were made from the original file:
# 1. Retrieved from https://github.com/batfish/pybatfish/blob/e23829854c70446b4e68a3f8ef2a25410daad167/pybatfish/util.py
# 2. All code except function zip_dir, its imports, and constant _MIN_ZIP_TIMESTAMP was excluded

import os
import tempfile
from typing import IO
import zipfile

# Minimum timestamp supported by ZIP format
# See issue https://bugs.python.org/issue34097
_MIN_ZIP_TIMESTAMP = 315561600.0


def zip_dir(dir_path: str, out_file: str | IO):
    """
    ZIP a specified directory and write it to the given output file path.

    :param dir_path: path to the directory to be zipped up
    :type dir_path: str
    :param out_file: path to the resulting zipfile
    :type out_file: str
    """
    with zipfile.ZipFile(out_file, "w", zipfile.ZIP_DEFLATED) as zipWriter:
        rel_root = os.path.abspath(os.path.join(dir_path, os.path.pardir))

        for root, _dirs, files in os.walk(dir_path):
            zipWriter.write(root, os.path.relpath(root, rel_root), zipfile.ZIP_STORED)
            for f in files:
                filename = os.path.join(root, f)
                arcname = os.path.join(os.path.relpath(root, rel_root), f)

                # Zipped files must be from 1980 or later
                # So copy any file older than that to a tempfile to bump the timestamp
                if os.path.getmtime(filename) < _MIN_ZIP_TIMESTAMP:
                    with tempfile.NamedTemporaryFile("w+b") as temp_file, open(
                        filename, "rb"
                    ) as file_src:
                        temp_file.write(file_src.read())
                        temp_file.flush()
                        zipWriter.write(temp_file.name, arcname)
                else:
                    zipWriter.write(filename, arcname)



from typing import Any, Dict, Type, TypeVar

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

from ..types import UNSET, Unset
from typing import cast, Union
from typing import Literal, Union


T = TypeVar("T", bound="ConsumeClientLoginSessionResponse")


@_attrs_define
class ConsumeClientLoginSessionResponse:
    """
    Attributes:
        status (int):
        retry_after (int):
        access_token (Union[None, str]):
        org_name (Union[None, str]):
        type (Union[Literal['urn:invariant:responses:consume_client_login_response'], Unset]):  Default:
            'urn:invariant:responses:consume_client_login_response'.
    """

    status: int
    retry_after: int
    access_token: Union[None, str]
    org_name: Union[None, str]
    type: Union[
        Literal["urn:invariant:responses:consume_client_login_response"], Unset
    ] = "urn:invariant:responses:consume_client_login_response"
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        status = self.status
        retry_after = self.retry_after
        access_token: Union[None, str]

        access_token = self.access_token

        org_name: Union[None, str]

        org_name = self.org_name

        type = self.type

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "status": status,
                "retry_after": retry_after,
                "access_token": access_token,
                "org_name": org_name,
            }
        )
        if type is not UNSET:
            field_dict["type"] = type

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        status = d.pop("status")

        retry_after = d.pop("retry_after")

        def _parse_access_token(data: object) -> Union[None, str]:
            if data is None:
                return data
            return cast(Union[None, str], data)

        access_token = _parse_access_token(d.pop("access_token"))

        def _parse_org_name(data: object) -> Union[None, str]:
            if data is None:
                return data
            return cast(Union[None, str], data)

        org_name = _parse_org_name(d.pop("org_name"))

        type = d.pop("type", UNSET)

        consume_client_login_session_response = cls(
            status=status,
            retry_after=retry_after,
            access_token=access_token,
            org_name=org_name,
            type=type,
        )

        consume_client_login_session_response.additional_properties = d
        return consume_client_login_session_response

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, Type, TypeVar

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

from ..types import UNSET, Unset
from typing import cast, Union
from typing import Union


T = TypeVar("T", bound="ReportTextSummaryResponse")


@_attrs_define
class ReportTextSummaryResponse:
    """
    Attributes:
        text (str):
        json (Union[None, Unset, str]):
    """

    text: str
    json: Union[None, Unset, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        text = self.text
        json: Union[None, Unset, str]
        if isinstance(self.json, Unset):
            json = UNSET

        else:
            json = self.json

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "text": text,
            }
        )
        if json is not UNSET:
            field_dict["json"] = json

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        text = d.pop("text")

        def _parse_json(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        json = _parse_json(d.pop("json", UNSET))

        report_text_summary_response = cls(
            text=text,
            json=json,
        )

        report_text_summary_response.additional_properties = d
        return report_text_summary_response

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import argparse
import datetime
import io
import json
import logging
import logging.config
import os
import pathlib
from retry.api import retry_call
import ssl
import stat
import sys
import time
import uuid

from attrs import asdict
from tabulate import tabulate

from invariant_client import auth, display, zip_util
from invariant_client import pysdk
from invariant_client.bindings.invariant_instance_client.models.snapshot_report_data import SnapshotReportData
from invariant_client.display import OutputFormat


CREDS_FILE_PATH = pathlib.Path.cwd()
try:
    CREDS_FILE_PATH = pathlib.Path.home()
except RuntimeError:
    pass
finally:
    CREDS_FILE_PATH = CREDS_FILE_PATH.joinpath('.invariant_creds')

class UploadTerminationError(Exception):
    """An exception that is raised when a snapshot upload is terminated."""

def parse_args():
    parser = argparse.ArgumentParser(
        prog='invariant',
        description='Invariant analyzes network snapshots',
        formatter_class=argparse.RawTextHelpFormatter,
    )
    subparsers = parser.add_subparsers(
        title='available commands',
        description='Run [command] --help for more information.',
        dest='command')
    command_login       = subparsers.add_parser(
        'login',
        description='Authenticate by opening a link in your browser. Saves credentials to ~/.invariant_creds.',
        help="Authenticate by opening a link in your browser. Saves credentials to ~/.invariant_creds.")
    command_run         = subparsers.add_parser(
        'run',
        description='Analyze the current directory.',
        help="Analyze the current directory.")
    command_show        = subparsers.add_parser(
        'show',
        description='Examine Invariant analysis results.',
        help="Examine Invariant analysis results.")
    command_solution    = subparsers.add_parser(
        'show_solution',
        description='Display a suggested patch result.',
        help="Display a suggested patch result.")
    command_snapshots   = subparsers.add_parser(
        'snapshots',
        description='List prior snapshot analysis results.',
        help="List prior snapshot analysis results.")

    def add_common_arguments(parser):
        parser.add_argument(
            '--json',
            dest='json',
            action='store_true',
            help='Output data as JSON.',
        )
        parser.add_argument(
            '--tsv',
            dest='tsv',
            action='store_true',
            help='Output data as TSV.',
        )
        parser.add_argument(
            '--debug',
            dest='debug',
            action='store_true',
            help='Enable detailed logging.',
        )
    
    add_common_arguments(command_login)
    add_common_arguments(command_run)
    add_common_arguments(command_snapshots)
    add_common_arguments(command_show)
    add_common_arguments(command_solution)

    command_run.add_argument(
        '--compare-to',
        dest='compare_to',
        help='Compare this snapshot to another by its git ref. Ref must refer to the primary git repository.',
    )

    command_run.add_argument(
        '--target',
        dest='target',
        help='An Invariant project root directory. Default is current directory.',
    )

    command_show.add_argument(
        'snapshot_name',
        help='The snapshot to examine.'
    )

    command_show.add_argument(
        'file_name',
        nargs="?",
        help='The snapshot file to examine.'
    )

    command_solution.add_argument(
        'snapshot_name',
        help='The snapshot to examine.'
    )

    command_solution.add_argument(
        'solution_name',
        help='The solution to examine.'
    )

    args = parser.parse_args()

    command = getattr(args, 'command')
    if not command:
        parser.print_help()
        exit(0)

    return args


def serialize(inst, field, value):
    if isinstance(value, datetime.datetime):
        return value.isoformat()
    return value


def configure_logging(debug: bool):
    if debug:
        logging.basicConfig(level=logging.DEBUG)
        logging.config.dictConfig(
            {
                "version": 1,
                "disable_existing_loggers": False,
                "incremental": True,
                "loggers": {
                    "multipart.multipart": {"level": "INFO"},
                },
            }
        )
    try:
        import rich.console
        import rich.logging
        root_logger = logging.getLogger()
        rich_handler = rich.logging.RichHandler(
            rich_tracebacks=True,
            omit_repeated_times=False,
            tracebacks_show_locals=True,
            tracebacks_suppress=[],
            show_time=False,
            show_path=False,
            console=rich.console.Console(stderr=True),
        )
        root_logger.addHandler(rich_handler)
    except ImportError:
        pass

    
def EntryPoint():
    args = parse_args()

    command = getattr(args, 'command')

    format = OutputFormat.TABULATE
    if getattr(args, 'json'):
        format = OutputFormat.JSON
    elif getattr(args, 'tsv'):
        format = OutputFormat.TSV
    
    debug = getattr(args, 'debug') or False
    configure_logging(debug)
    
    try:
        EntryPoint_inner(args, command, format, debug)
    except Exception as e:
        if debug:
            raise e
        print('Error: %s' % e, file=sys.stderr)
        exit(1)


def EntryPoint_inner(args, command, format, debug):
    settings: pysdk.Settings = {
        'format': format,
        'debug': debug,
    }

    env = dict(os.environ)
    invariant_domain = env.get('INVARIANT_DOMAIN', 'https://invariant.tech')

    creds = None

    if command == 'login':
        # TODO warn before logging in if an API token is present (possibly check if it works?)
        workflow = auth.BrowserLoginFlow(invariant_domain, ssl.create_default_context())
        link = workflow.start()
        print("Open this link in your browser to log in:")
        print(link)
        try:
            end_time = datetime.datetime.now() + datetime.timedelta(minutes=3)
            # time.sleep(10)  # poor man's websocket
            time.sleep(6)  # poor man's websocket
            # TODO consider a nice animated "waiting" message for interactive terminal
            while not creds and end_time > datetime.datetime.now():
                result = workflow.poll_await_browser_creds()
                if isinstance(result, pysdk.AccessCredential):
                    creds = result
                    break
                elif isinstance(result, int):
                    time.sleep(result)
                else:
                    time.sleep(2)
            if not creds:
                print("Timed out.")
                exit(1)
            with open(CREDS_FILE_PATH, 'w') as f:
                f.write(creds.to_json())
            CREDS_FILE_PATH.chmod(
                stat.S_IRUSR |
                stat.S_IWUSR
            )
            print("Login successful.")
        except KeyboardInterrupt as e:
            print("Exiting...")
            exit(1)

        exit(0)
    
    # Load credentials or error
    try:
        creds = pysdk.AccessCredential.from_env(env, base_url=invariant_domain)
        if not creds:
            try:
                creds = pysdk.AccessCredential.from_file(CREDS_FILE_PATH, base_url=invariant_domain)
            except FileNotFoundError:
                # Expected
                creds = None
            if not creds:
                print("Please run 'invariant login' to authenticate.", file=sys.stderr)
                exit(1)

    except pysdk.AuthorizationException as e:
        print(f"Error: {e}", file=sys.stderr)
        print("Please run 'invariant login' to authenticate.", file=sys.stderr)
        if debug:
            raise e
        exit(1)
    except pysdk.RemoteError as e:
        print(f"Error: {e}", file=sys.stderr)
        if debug:
            raise e
        exit(1)

    sdk = pysdk.Invariant(
        creds=creds,
        settings=settings,
        base_url=invariant_domain)

    if command == "run":
        target = getattr(args, 'target') or '.'
        bytes = None
        if pathlib.Path(target).is_file():
            with open(target, "rb") as f:
                bytes = io.BytesIO(f.read())
        elif pathlib.Path(target).is_dir():
            bytes = io.BytesIO()
            zip_util.zip_dir(target, bytes)  # Write a zip file into 'bytes'
        else:
            print("Unacceptable target", file=sys.stderr)
            print(str(target), file=sys.stderr)
            exit(1)

        compare_to = getattr(args, 'compare_to')
        exec_uuid = retry_call(upload_snapshot, fargs=[sdk, bytes, compare_to], exceptions=UploadTerminationError,tries=3, delay=30, backoff=2)
        print("Analysis complete.")
        response = sdk.snapshot_detail(exec_uuid)
        # pprint.pprint(asdict(response), width=200)
        display.snapshot_status(response)
        if response.status['state'] == 'COMPLETE':
            display.snapshot_halted(response)
            print('')
            display.snapshot_summary_table(response, format)
            print(f"\nRun 'invariant show {exec_uuid} <file>' to examine any file.")

            if response.summary['errors'] > 0:
                print(f"\n{response.summary['errors']} {'error' if response.summary['errors'] == 1 else 'errors'} found.")
                errors_uuid = response.report.reports.errors
                errors_response = sdk.snapshot_file(errors_uuid)
                display.snapshot_errors(errors_response, format)

        else:
            if response.summary['errors'] > 0:
                errors_uuid = response.report.reports.errors
                errors_response = sdk.snapshot_file(errors_uuid)
                display.snapshot_errors(errors_response, format)

    elif command == "snapshots":
        snapshots = sdk.list_snapshots()
        if format == OutputFormat.JSON:
            print(json.dumps(asdict(snapshots, value_serializer=serialize), default=vars))
        elif format == OutputFormat.TSV:
            reports = asdict(snapshots)
            reports = reports['reports']
            print(tabulate(reports, headers='keys', tablefmt='tsv'))
        # if format == 'markdown':
        #     return tabulate(result, headers='keys', tablefmt='github')
        else:
            reports = asdict(snapshots)
            reports = reports['reports']
            print(tabulate(reports, headers='keys', tablefmt='psql'))

    elif command == "snapshot":
        sdk.snapshot_detail(
            report_uuid=args.snapshot_name)

    elif command == "show":
        try:
            exec_uuid = uuid.UUID(args.snapshot_name, version=4)
        except ValueError as e:
            # TODO we could permit something like 'latest'
            raise ValueError(f"Expected {args.snapshot_name} to be a UUID like f5b4e387-e336-499e-b3a0-d6186c590572.") from e

        if args.file_name is not None:
            # Access a specific file
            try:
                file = uuid.UUID(args.file_name, version=4)
            except ValueError:
                # OK if the file is the file key (e.g. errors)
                file = args.file_name
            if not isinstance(file, uuid.UUID):
                # Resolve non-UUID file to UUID
                response = sdk.snapshot_detail(exec_uuid)
                reports = response.report.reports
                try:
                    file: str = getattr(reports, file)
                    file = uuid.UUID(file, version=4)
                except AttributeError as e:
                    if not isinstance(reports, SnapshotReportData):
                        raise ValueError(f"Report {file} not found for snapshot {exec_uuid}.") from e
                    try:
                        file: str = reports.files[file]
                        file = uuid.UUID(file, version=4)
                    except KeyError as e:
                        raise ValueError(f"Report {file} not found for snapshot {exec_uuid}.") from e

            if format == OutputFormat.JSON:
                file_summary = sdk.snapshot_file_text(str(file), False, json_mode=True)
                if file_summary.json:
                    print(file_summary.json)
                else:
                    file_data = sdk.snapshot_file(str(file))
                    print(file_data.to_json(orient='records'))
            elif format == OutputFormat.TSV:
                file_data = sdk.snapshot_file(str(file))
                display.print_frame(file_data, format)
            else:
                file_summary = sdk.snapshot_file_text(str(file), False, json_mode=False)
                if file_summary.text:
                    print(file_summary.text)
                else:
                    file_data = sdk.snapshot_file(str(file))
                    display.print_frame(file_data, format)
                # print("Set --traces to display all example traces")
                print("Set --json to get JSON")
                print("See 'show --help' for more options")

        else:
            # Display the process summary for the snapshot
            response = sdk.snapshot_detail(exec_uuid)
            display.snapshot_status(response)
            if response.status['state'] == 'COMPLETE':
                display.snapshot_halted(response)
                print('')
                display.snapshot_summary_table(response, format)
                print(f"\nRun 'invariant show {exec_uuid} <file>' to examine any file.")

                if response.summary['errors'] > 0:
                    print(f"\n{response.summary['errors']} {'error' if response.summary['errors'] == 1 else 'errors'} found.")
                    errors_uuid = response.report.reports.errors
                    errors_response = sdk.snapshot_file(errors_uuid)
                    display.snapshot_errors(errors_response, format)

            else:
                if response.summary['errors'] > 0:
                    errors_uuid = response.report.reports.errors
                    errors_response = sdk.snapshot_file(errors_uuid)
                    display.snapshot_errors(errors_response, format)

    elif command == "show_solution":
        sdk.show_solution(
            snapshot=args.snapshot_name,
            solution=args.solution_name)

    else:
        print(f"Unknown command {command}", file=sys.stderr)


def upload_snapshot(sdk: pysdk.Invariant, bytes: io.BytesIO, compare_to: str) -> str:
    print("Uploading snapshot...")
    exec_uuid = sdk.upload_snapshot(
        source=bytes,
        compare_to=compare_to)
    exec_uuid = exec_uuid.exec_uuid
    end_time = datetime.datetime.now() + datetime.timedelta(weeks=1)
    print(f"Processing... ({exec_uuid})")
    while datetime.datetime.now() < end_time:
        response = sdk.upload_is_running(exec_uuid)
        if response.terminated:
            raise UploadTerminationError(f"Upload was terminated")
        if not response.is_running:
            break
        
        # TODO send some RetryAfter header to control this
        # TODO separately, exponential back-off on error
        time.sleep(4)
    if not response:
        print("Timed out.")
        exit(1)
    return exec_uuid


from http import HTTPStatus
from typing import Any, Dict, List, Optional, Union

import httpx

from ...client import AuthenticatedClient, Client
from ...types import Response
from ... import errors

from typing import List
from ...models.validation_error_response import ValidationErrorResponse
from ...models.challenge_response import ChallengeResponse
from ...models.user import User
from typing import Dict


def _get_kwargs() -> Dict[str, Any]:
    return {
        "method": "post",
        "url": "/api/v1/login/get_instances",
    }


def _parse_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Optional[Union[ChallengeResponse, List["User"], ValidationErrorResponse]]:
    if response.status_code == HTTPStatus.OK:
        response_200 = []
        _response_200 = response.json()
        for response_200_item_data in _response_200:
            response_200_item = User.from_dict(response_200_item_data)

            response_200.append(response_200_item)

        return response_200
    if response.status_code == HTTPStatus.UNPROCESSABLE_ENTITY:
        response_422 = ValidationErrorResponse.from_dict(response.json())

        return response_422
    if response.status_code == HTTPStatus.UNAUTHORIZED:
        response_401 = ChallengeResponse.from_dict(response.json())

        return response_401
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[Union[ChallengeResponse, List["User"], ValidationErrorResponse]]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    *,
    client: AuthenticatedClient,
) -> Response[Union[ChallengeResponse, List["User"], ValidationErrorResponse]]:
    """List instances available to the current login

     List instances where this login is associated with an active user.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[ChallengeResponse, List['User'], ValidationErrorResponse]]
    """

    kwargs = _get_kwargs()

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    *,
    client: AuthenticatedClient,
) -> Optional[Union[ChallengeResponse, List["User"], ValidationErrorResponse]]:
    """List instances available to the current login

     List instances where this login is associated with an active user.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[ChallengeResponse, List['User'], ValidationErrorResponse]
    """

    return sync_detailed(
        client=client,
    ).parsed


async def asyncio_detailed(
    *,
    client: AuthenticatedClient,
) -> Response[Union[ChallengeResponse, List["User"], ValidationErrorResponse]]:
    """List instances available to the current login

     List instances where this login is associated with an active user.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[ChallengeResponse, List['User'], ValidationErrorResponse]]
    """

    kwargs = _get_kwargs()

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    *,
    client: AuthenticatedClient,
) -> Optional[Union[ChallengeResponse, List["User"], ValidationErrorResponse]]:
    """List instances available to the current login

     List instances where this login is associated with an active user.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[ChallengeResponse, List['User'], ValidationErrorResponse]
    """

    return (
        await asyncio_detailed(
            client=client,
        )
    ).parsed



from typing import Any, Dict, Type, TypeVar, TYPE_CHECKING

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

from typing import Union
from typing import cast, Union
from typing import cast
from ..types import UNSET, Unset
from typing import Dict

if TYPE_CHECKING:
    from ..models.comparison_reportdata import ComparisonReportdata
    from ..models.snapshot_report_data_files import SnapshotReportDataFiles


T = TypeVar("T", bound="SnapshotReportData")


@_attrs_define
class SnapshotReportData:
    """
    Attributes:
        files (SnapshotReportDataFiles):
        summary (str):
        status (str):
        errors (str):
        solutions (Union[None, Unset, str]):
        compare_to (Union['ComparisonReportdata', None, Unset]):
    """

    files: "SnapshotReportDataFiles"
    summary: str
    status: str
    errors: str
    solutions: Union[None, Unset, str] = UNSET
    compare_to: Union["ComparisonReportdata", None, Unset] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        from ..models.comparison_reportdata import ComparisonReportdata

        files = self.files.to_dict()

        summary = self.summary
        status = self.status
        errors = self.errors
        solutions: Union[None, Unset, str]
        if isinstance(self.solutions, Unset):
            solutions = UNSET

        else:
            solutions = self.solutions

        compare_to: Union[Dict[str, Any], None, Unset]
        if isinstance(self.compare_to, Unset):
            compare_to = UNSET

        elif isinstance(self.compare_to, ComparisonReportdata):
            compare_to = UNSET
            if not isinstance(self.compare_to, Unset):
                compare_to = self.compare_to.to_dict()

        else:
            compare_to = self.compare_to

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "files": files,
                "summary": summary,
                "status": status,
                "errors": errors,
            }
        )
        if solutions is not UNSET:
            field_dict["solutions"] = solutions
        if compare_to is not UNSET:
            field_dict["compare_to"] = compare_to

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.comparison_reportdata import ComparisonReportdata
        from ..models.snapshot_report_data_files import SnapshotReportDataFiles

        d = src_dict.copy()
        files = SnapshotReportDataFiles.from_dict(d.pop("files"))

        summary = d.pop("summary")

        status = d.pop("status")

        errors = d.pop("errors")

        def _parse_solutions(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        solutions = _parse_solutions(d.pop("solutions", UNSET))

        def _parse_compare_to(
            data: object,
        ) -> Union["ComparisonReportdata", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                _compare_to_type_0 = data
                compare_to_type_0: Union[Unset, ComparisonReportdata]
                if isinstance(_compare_to_type_0, Unset):
                    compare_to_type_0 = UNSET
                else:
                    compare_to_type_0 = ComparisonReportdata.from_dict(
                        _compare_to_type_0
                    )

                return compare_to_type_0
            except:  # noqa: E722
                pass
            return cast(Union["ComparisonReportdata", None, Unset], data)

        compare_to = _parse_compare_to(d.pop("compare_to", UNSET))

        snapshot_report_data = cls(
            files=files,
            summary=summary,
            status=status,
            errors=errors,
            solutions=solutions,
            compare_to=compare_to,
        )

        snapshot_report_data.additional_properties = d
        return snapshot_report_data

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, Type, TypeVar, TYPE_CHECKING

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field


from typing import Dict
from typing import cast
from typing import cast, List

if TYPE_CHECKING:
    from ..models.integration_data_github_app_installation_data_extra import (
        IntegrationDataGithubAppInstallationDataExtra,
    )


T = TypeVar("T", bound="IntegrationDataGithubAppInstallationData")


@_attrs_define
class IntegrationDataGithubAppInstallationData:
    """
    Attributes:
        user_creator (str):
        gh_creator (str):
        installation_id (str):
        target_orgs (List[str]):
        target_repos (List[str]):
        extra (IntegrationDataGithubAppInstallationDataExtra):
    """

    user_creator: str
    gh_creator: str
    installation_id: str
    target_orgs: List[str]
    target_repos: List[str]
    extra: "IntegrationDataGithubAppInstallationDataExtra"
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        user_creator = self.user_creator
        gh_creator = self.gh_creator
        installation_id = self.installation_id
        target_orgs = self.target_orgs

        target_repos = self.target_repos

        extra = self.extra.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "user_creator": user_creator,
                "gh_creator": gh_creator,
                "installation_id": installation_id,
                "target_orgs": target_orgs,
                "target_repos": target_repos,
                "extra": extra,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.integration_data_github_app_installation_data_extra import (
            IntegrationDataGithubAppInstallationDataExtra,
        )

        d = src_dict.copy()
        user_creator = d.pop("user_creator")

        gh_creator = d.pop("gh_creator")

        installation_id = d.pop("installation_id")

        target_orgs = cast(List[str], d.pop("target_orgs"))

        target_repos = cast(List[str], d.pop("target_repos"))

        extra = IntegrationDataGithubAppInstallationDataExtra.from_dict(d.pop("extra"))

        integration_data_github_app_installation_data = cls(
            user_creator=user_creator,
            gh_creator=gh_creator,
            installation_id=installation_id,
            target_orgs=target_orgs,
            target_repos=target_repos,
            extra=extra,
        )

        integration_data_github_app_installation_data.additional_properties = d
        return integration_data_github_app_installation_data

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



# Invariant client authentication library


from dataclasses import dataclass
import enum
import ssl
from typing import Optional
from invariant_client import pysdk

from invariant_client.pysdk import AccessCredential, RemoteError

from invariant_client.bindings.invariant_login_client import models
from invariant_client.bindings.invariant_login_client.client import AuthenticatedClient as LoginAuthenticatedClient, Client as LoginClient
from invariant_client.bindings.invariant_login_client.api.client_login.init_client_login_api_v1_client_login_post import sync_detailed as init_client_login_api_v1_client_login_post
from invariant_client.bindings.invariant_login_client.api.client_login.consume_client_login_api_v1_client_login_consume_post import sync_detailed as consume_client_login_api_v1_client_login_consume_post


class Unauthorized(Exception):
    """Credentials are not valid."""


class BrowserLoginFlowState(enum.Enum):
    START = enum.auto()
    AWAIT_BROWSER = enum.auto()
    COMPLETE = enum.auto()


class BrowserLoginFlow:
    """ In the browser login flow, the client initiates the login flow, recieves a code (or URL),
    sends the user to the Invariant website in their browser to complete the flow, then
    polls for a response from the server. The server response will contain the CLI
    access control credentials."""

    state: BrowserLoginFlowState
    client_base_url: str
    client_kwargs: dict
    pin_url: Optional[str]         # The PIN code is shared with the user
    client_token: Optional[str]   # The client session token can consume the client login session once fulfilled by the user
    error: Optional[Exception]
    creds: Optional[AccessCredential]   # Successfully gained access_token creds go here

    def __init__(self,
            base_url: Optional[str] = None,
            verify_ssl: Optional[str | bool | ssl.SSLContext] = None,
            **kwargs,
        ):
        self.client_base_url = base_url or pysdk.DOMAIN_NAME
        self.client_kwargs = {}
        self.client_kwargs.update(kwargs)
        self.client_kwargs['verify_ssl'] = verify_ssl
        self.state = BrowserLoginFlowState.START
        self.pin_url = None
        self.client_token = None
        self.error = None
        self.creds = None

    def start(self) -> str:
        """Connect to invariant.tech and initiate a browser login flow.
        
        Returns an external login code (string). The user should use this code to complete the login flow
        in their browser."""
        client = LoginClient(self.client_base_url, **self.client_kwargs)
        response = init_client_login_api_v1_client_login_post(client=client)
        response = response.parsed
        if not response:
            raise RemoteError(f"Unable to connect to {self.client_base_url}")
        if not isinstance(response, models.CreateClientLoginSessionResponse):
            raise RemoteError(response)
        self.state = BrowserLoginFlowState.AWAIT_BROWSER
        self.pin_url = response.url
        self.client_token = response.token
        return self.pin_url

    def poll_await_browser_creds(self) -> Optional[AccessCredential | str]:
        client = LoginAuthenticatedClient(
            self.client_base_url,
            token=self.client_token,
            **self.client_kwargs)
        response = consume_client_login_api_v1_client_login_consume_post(client=client)
        response = response.parsed
        if not response:
            raise RemoteError(f"Unable to connect to {self.client_base_url}")
        if not isinstance(response, models.ConsumeClientLoginSessionResponse):
            raise RemoteError(response)

        if response.access_token:
            self.creds = AccessCredential(
                access_token=response.access_token,
                refresh_token=None,
                organization_name=response.org_name
            )
            self.state = BrowserLoginFlowState.COMPLETE
            return self.creds
        elif response.retry_after:
            return int(response.retry_after)
        else:
            return 10


    @property
    def success(self):
        return self.state == BrowserLoginFlowState.COMPLETE and self.error is None

